#=======================================================================
# Bits_test.py
#=======================================================================
# Tests for the Bits class.
# Shunning: grabbed from PyMTL2. Thanks Derek Lockhart

from copy import deepcopy

import pytest

from ..bits_import import Bits


def test_return_type():

  x = Bits( 8, 0b1100 )
  assert isinstance( x.uint(), int  )
  assert isinstance( x.int(),  int  )
  assert isinstance( x[1:2],   Bits )
  assert isinstance( x[0:4],   Bits )
  assert isinstance( x[2],     Bits )

def test_int():

  assert Bits( 4,  0 ).int() == 0
  assert Bits( 4,  2 ).int() == 2
  assert Bits( 4,  4 ).int() == 4
  assert Bits( 4, 15 ).int() == -1
  assert Bits( 4, -1 ).int() == -1
  assert Bits( 4, -2 ).int() == -2
  assert Bits( 4, -4 ).int() == -4
  assert Bits( 4, -8 ).int() == -8

def test_int_bounds_checking():

  Bits( 4, 15 )
  with pytest.raises( ValueError ): Bits( 4, 16 )

  Bits( 4, -8 )
  with pytest.raises( ValueError ): Bits( 4, -9 )
  with pytest.raises( ValueError ): Bits( 4, -16 )

  Bits( 1, 0 )
  Bits( 1, 1 )
  Bits( 1, -1 )
  # with pytest.raises( ValueError ): Bits( 1, -1 )
  with pytest.raises( ValueError ): Bits( 1, -2 )

def test_uint():

  assert Bits( 4,  0 ).uint() == 0
  assert Bits( 4,  2 ).uint() == 2
  assert Bits( 4,  4 ).uint() == 4
  assert Bits( 4, 15 ).uint() == 15
  assert Bits( 4, -1 ).uint() == 15
  assert Bits( 4, -2 ).uint() == 14
  assert Bits( 4, -4 ).uint() == 12

def test_neg_assign():

  x = Bits( 4, -1 )
  assert x        == 0b1111
  assert x.uint() == 0b1111
  x = Bits( 4, -2 )
  assert x        == 0b1110
  assert x.uint() == 0b1110

def test_get_bit():

  x = Bits( 4, 0b1100 )
  assert x[3] == 1
  assert x[2] == 1
  assert x[1] == 0
  assert x[0] == 0

def test_set_bit():

  x = Bits( 4, 0b1100 )
  x[3] = 0
  assert x.uint() == 0b0100
  x[2] = 1
  assert x.uint() == 0b0100
  x[1] = 1
  assert x.uint() == 0b0110

  x = Bits( 4, 0b100 )
  x[2] = Bits(1,0)
  assert x.uint() == 0

  with pytest.raises( ValueError ):
    x[2] = Bits(2,1)

def test_bit_bounds_checking():

  x = Bits( 4, 0b1100 )
  with pytest.raises( IndexError ):
    assert x[-1] == 1
  with pytest.raises( IndexError ):
    assert x[8] == 1
  with pytest.raises( IndexError ):
    x[-1] = 1
  with pytest.raises( IndexError ):
    x[4] = 1
  with pytest.raises( ValueError ):
    x[0] = 2
  # with pytest.raises( ValueError ):
  x[3] = -1

def test_get_slice():

  x = Bits( 4, 0b1100 )
  assert x[:] == 0b1100
  assert x[2:4] == 0b11
  assert x[0:1] == 0b0
  assert x[1:3] == 0b10
  # check open ended ranges
  assert x[1:] == 0b110
  assert x[:3] == 0b100

  with pytest.raises( IndexError ):
    x[1:3:1]

def test_set_slice():

  x = Bits( 4, 0b1100 )
  x[:] = 0b0010
  assert x.uint() == 0b0010
  x[2:4] = 0b11
  assert x.uint() == 0b1110
  x[0:1] = 0b1
  assert x.uint() == 0b1111
  x[1:3] = 0b10
  assert x.uint() == 0b1101
  # check open ended ranges
  x[1:] = 0b001
  assert x.uint() == 0b0011
  x[:3] = 0b110
  assert x.uint() == 0b0110

  with pytest.raises( ValueError ):
    x[1:3] = Bits(1,1)

  with pytest.raises( ValueError ):
    x[1:3] = Bits(3,1)

  x[1:3] = Bits(2,1)
  assert x.uint() == 0b0010

  x[:]   = 0b1111
  assert x.uint() == 0b1111

  with pytest.raises( ValueError ):
    x[:]   = 0b10000

  with pytest.raises( IndexError ):
    x[1:4:2] = 1

def test_slice_bounds_checking():

  x = Bits( 4, 0b1100 )
  with pytest.raises( IndexError ):
    assert x[1:5]  == 0b10
  with pytest.raises( IndexError ):
    assert x[-1:2] == 0b10
  with pytest.raises( IndexError ):
    assert x[2:1]  == 0b10
  with pytest.raises( IndexError ):
    x[1:5]  = 0b10
  with pytest.raises( IndexError ):
    x[-1:2] = 0b10
  with pytest.raises( IndexError ):
    x[2:1]  = 0b10
  # FIXED
  # Bits objects constructed with another Bits object provided as a value
  # parameter end up having problems when writing to slices.  This is
  # because the mask used when writing to a subslice is often a negative
  # int in Python, and we don't allow operations on Bits to be performed
  # with negative values.  Current workaround is to force the value param
  # for the Bits constructor to be an int or a long.
  #with pytest.raises( AssertionError ):
  y = Bits( 4, Bits( 4, 0 ))
  y[1:3] = 1


def test_eq():

  x = Bits( 4, 0b1010 )
  assert x.uint() == x.uint()
  # Compare objects by value, not id
  assert x == x
  # Check the value
  assert x.uint() == 0b1010
  assert x.uint() == 0xA
  assert x.uint() == 10
  # Checking the equality operator
  assert x == 0b1010
  assert x == 0xA
  assert x == 10
  # Checking comparison with another bit container
  y = Bits( 4, 0b1010 )
  assert x.uint() == y.uint()
  assert x == y
  y = Bits( 8 , 0b1010 )
  assert x.uint() == y.uint()
  # TODO: How should equality between Bits objects work?
  #       Just same value or same value and width?
  #assert x == y
  # Check the negatives
  x = Bits( 4, -1 )
  assert x.uint() == 0b1111
  assert x.uint() == 0xF
  assert x.uint() == 15
  # Checking the equality operator
  assert x == 0b1111
  assert x == 0xF
  assert x == 15
  assert x.uint() == Bits(4, -1).uint()
  assert x == Bits( 4, -1 ).uint()
  assert 15 == x

  assert not x == None
  assert not Bits( 4, 0 ) == None
  with pytest.raises( ValueError ):
    Bits(4,3) == Bits(3, 2)

def test_ne():

  x = Bits( 4, 0b1100 )
  y = Bits( 4, 0b0011 )
  # TODO: check width?
  assert x.uint() != y.uint()
  assert x != y
  # added for bug
  z = Bits( 1, 0 )
  assert z.uint() != 1
  assert z != 1
  assert 5 != x

  assert z != None
  assert Bits( 4, 0 ) != None

def test_compare_neg_assert():

  x = Bits( 4, -2 )
  # We don't allow comparison with negative numbers,
  # although you can construct a new Bits object with one...
  with pytest.raises( ValueError ):
    assert x != -1
  with pytest.raises( ValueError ):
    assert x == -2
  with pytest.raises( ValueError ):
    assert x >  -3
  with pytest.raises( ValueError ):
    assert x >= -3
  with pytest.raises( ValueError ):
    assert x <  -1
  with pytest.raises( ValueError ):
    assert x >= -1
  assert x != Bits( 4, -1 )
  assert x == Bits( 4, -2 )
  assert x >  Bits( 4, -3 )
  assert x >= Bits( 4, -3 )
  assert x <  Bits( 4, -1 )
  assert x <= Bits( 4, -1 )

def test_compare_uint_neg():

  x = Bits( 4, 2 )
  assert x.uint() != -1
  assert x.uint()  > -1
  assert x.uint() >= -1

def test_compare_int_neg():

  x = Bits( 4, -2 )
  assert x.int() == -2
  assert x.int()  < -1
  assert x.int() <= -1

def test_lt():

  x = Bits( 4, 0b1100 )
  y = Bits( 4, 0b0011 )
  assert y.uint() < x.uint()
  assert y.uint() < 10
  assert y < x.uint()
  assert y < 10
  assert y < x
  assert 1 < y
  with pytest.raises( ValueError ):
    x < -1
  with pytest.raises( ValueError ):
    Bits(4,3) < Bits(3, 2)

def test_gt():

  x = Bits( 4, 0b1100 )
  y = Bits( 4, 0b0011 )
  assert x.uint() > y.uint()
  assert x.uint() > 2
  assert x > y.uint()
  assert x > 2
  assert x > y
  assert 9 > y
  with pytest.raises( ValueError ):
    x > -1
  with pytest.raises( ValueError ):
    Bits(4,3) > Bits(3, 2)

def test_lte():

  x = Bits( 4, 0b1100 )
  y = Bits( 4, 0b0011 )
  z = Bits( 4, 0b0011 )
  assert y.uint() <= x.uint()
  assert y.uint() <= 10
  assert y.uint() <= z.uint()
  assert y.uint() <= 0b0011
  assert y <= x.uint()
  assert y <= 10
  assert y <= z.uint()
  assert y <= 0b0011
  assert y <= x
  assert y <= z
  assert z <= x
  assert z <= z
  assert 1 <= y
  assert 3 <= y
  with pytest.raises( ValueError ):
    x <= -1
  with pytest.raises( ValueError ):
    Bits(4,3) <= Bits(3, 2)

def test_gte():

  x = Bits( 4, 0b1100 )
  y = Bits( 4, 0b0011 )
  z = Bits( 4, 0b1100 )
  assert x.uint() >= y.uint()
  assert x.uint() >= 2
  assert x.uint() >= z.uint()
  assert x.uint() >= 0b1100
  assert x >= y.uint()
  assert x >= 2
  assert x >= z.uint()
  assert x >= 0b1100
  assert x >= y
  assert x >= z
  assert z >= y
  assert z >= x
  assert x >= x
  assert 5 >= y
  assert 3 <= y
  with pytest.raises( ValueError ):
    x >= -1
  with pytest.raises( ValueError ):
    Bits(4,3) >= Bits(3, 2)

def test_invert():

  x = Bits( 4, 0b0001 )
  assert ~x == 0b1110
  x = Bits( 4, 0b1001 )
  assert ~x == 0b0110
  x = Bits( 16, 0b1111000011110000 )
  assert ~x == 0b0000111100001111

def test_add():

  x = Bits( 4, 4 )
  y = Bits( 4, 4 )
  assert x + y == 8
  assert x + Bits( 4, 4 ) == 8
  assert x + 4 == 8
  y = Bits( 4, 14 )
  assert x + y == 2
  assert x + 14 == 2
  assert 14 + x == 2

  a = Bits( 4, 1 )
  b = Bits( 4, 1 )
  c = Bits( 4, 1 )
  assert a + b + 1 == 3
  assert a + b + c == 3
  assert c + b + a == 3
  with pytest.raises( ValueError ):
    x + -1
  with pytest.raises( ValueError ):
    x + 100000000000000000000000000

  with pytest.raises( ValueError ):
    a = Bits(4,3) + Bits(3,1)

def test_sub():

  x = Bits( 4, 5 )
  y = Bits( 4, 4 )
  assert x - y == 1
  assert x - Bits(4, 4) == 1
  assert x - 4 == 1
  y = Bits( 4, 5 )
  assert x - y == 0
  assert x - 5 == 0
  y = Bits( 4, 7 )
  assert x - y == 0b1110
  assert x - 7 == 0b1110
  assert 9 - x == 0b0100
  with pytest.raises( ValueError ):
    x - -1
  with pytest.raises( ValueError ):
    x - 100000000000000000000000000

  with pytest.raises( ValueError ):
    a = Bits(4,3) - Bits(3,1)

def test_rsub():
  x = Bits(4, 5)
  y = 8
  z = y - x
  assert z.nbits == 4 and z.uint() == 3
  y = 16
  with pytest.raises( ValueError ):
    z = y - x

def test_floordiv():

  x = Bits( 4, 5 )
  y = Bits( 4, 4 )
  assert x // y == 1
  assert x // Bits(4, 4) == 1
  assert x // 4 == 1
  y = Bits( 4, 6 )
  assert x // y == 0
  assert x // 6 == 0
  assert 6 // x == 1
  with pytest.raises( ValueError ):
    x // -1
  with pytest.raises( ValueError ):
    x // 100000000000000000000000000

  with pytest.raises( ValueError ):
    a = Bits(4,3) // Bits(3,1)

def test_rfloordiv():
  x = Bits(4, 3)
  y = 8
  z = y // x
  assert z.nbits == 4 and z.uint() == 2
  y = 16
  with pytest.raises( ValueError ):
    z = y // x

def test_mod():

  x = Bits( 4, 5 )
  y = Bits( 4, 4 )
  assert x % y == 1
  assert x % Bits(4, 4) == 1
  assert x % 4 == 1
  y = Bits( 4, 6 )
  assert x % y == 5
  assert x % 6 == 5
  assert 6 % x == 1
  with pytest.raises( ValueError ):
    x % -1
  with pytest.raises( ValueError ):
    x % 100000000000000000000000000

  with pytest.raises( ValueError ):
    a = Bits(4,3) % Bits(3,1)

def test_rmod():
  x = Bits(4, 3)
  y = 8
  z = y % x
  assert z.nbits == 4 and z.uint() == 2
  y = 16
  with pytest.raises( ValueError ):
    z = y % x

def test_lshift():

  x = Bits( 8, 0b1100 )
  y = Bits( 8, 4 )
  assert x << y == 0b11000000
  assert x << 4 == 0b11000000
  assert x << 6 == 0b00000000
  assert y << x == 0b00000000
  assert y << 0 == 0b00000100
  assert y << 1 == 0b00001000
  assert x << 255 == 0

  with pytest.raises( ValueError ):
    a = Bits(4,3) << Bits(3,1)
  with pytest.raises( ValueError ):
    a = x << 256


def test_rshift():

  x = Bits( 8, 0b11000000 )
  y = Bits( 8, 4 )
  assert x >> y  == 0b00001100
  assert x >> 7  == 0b00000001
  assert x >> 8  == 0b00000000
  assert x >> 10 == 0b00000000
  x = Bits( 8, 2 )
  assert y >> x == 0b00000001
  assert y >> 0 == 0b00000100
  assert y >> 2 == 0b00000001
  assert y >> 5 == 0b00000000
  assert x >> 255 == 0

  with pytest.raises( ValueError ):
    a = Bits(4,3) >> Bits(3,1)
  with pytest.raises( ValueError ):
    a = x >> 256

def test_and():

  x = Bits( 8, 0b11001100 )
  y = Bits( 8, 0b11110000 )
  assert x & y      == 0b11000000
  assert x & 0b1010 == 0b00001000
  assert 0b1010 & x == 0b00001000
  with pytest.raises( ValueError ):
    x & -1
  with pytest.raises( ValueError ):
    x & 100000000000000000000000000

  with pytest.raises( ValueError ):
    a = Bits(4,3) & Bits(3,1)

def test_or():

  x = Bits( 8, 0b11001100 )
  y = Bits( 8, 0b11110000 )
  assert x | y      == 0b11111100
  assert x | 0b1010 == 0b11001110
  assert 0b1010 | x == 0b11001110
  with pytest.raises( ValueError ):
    x | -1
  with pytest.raises( ValueError ):
    x | 100000000000000000000000000
  with pytest.raises( ValueError ):
    a = Bits(4,3) | Bits(3,1)

def test_xor():

  x = Bits( 8, 0b11001100 )
  y = Bits( 8, 0b11110000 )
  assert x ^ y      == 0b00111100
  assert x ^ 0b1010 == 0b11000110
  assert 0b1010 ^ x == 0b11000110
  a = Bits( 1, 1 )
  b = Bits( 1, 0 )
  c = Bits( 1, 1 )
  assert ( a ^ b ) ^ c == 0
  with pytest.raises( ValueError ):
    x ^ -1
  with pytest.raises( ValueError ):
    x ^ 100000000000000000000000000
  with pytest.raises( ValueError ):
    a = Bits(4,3) ^ Bits(3,1)

# Now we always require the user to perform sext/zext of the multiply operand
def test_mult():

  x = Bits( 8, 0b00000000 )
  y = Bits( 8, 0b00000000 )
  assert x * y == 0b0000000000000000
  assert x * 0b1000 == 0b0000000000000000
  x = Bits( 8, 0b11111111 )
  y = Bits( 8, 0b11111111 )
  assert Bits( 16, x.uint() ) * Bits( 16, y.uint() ) == 0b0000000000000001111111000000001
  assert Bits( 16, x.uint() ) * 0b11111111 == 0b0000000000000001111111000000001
  assert 0b11111111 * Bits(16, x.uint() ) == 0b0000000000000001111111000000001

  # TODO: Currently fails as the second operand is larger than the Bits
  # object x. Should update the test when we define the behaviour
  #assert x * 0b1111111111 == 0b0000000000000001111111000000001

  y = Bits( 8, 0b10000000 )
  assert Bits( 16, x.uint() ) * Bits( 16, y.uint() ) == 0b0000000000000000111111110000000
  with pytest.raises( ValueError ):
    x * -1
  with pytest.raises( ValueError ):
    x * 100000000000000000000000000
  with pytest.raises( ValueError ):
    a = Bits(4,3) * Bits(3,1)

def test_constructor():

  assert Bits( 4,  2 ).uint() == 2
  assert Bits( 4,  4 ).uint() == 4
  assert Bits( 4, 15 ).uint() == 15
  assert Bits( 4, 17, trunc_int=True ).uint() == 1

  assert Bits( 4, -2 ).uint() == 0b1110
  assert Bits( 4, -4 ).uint() == 0b1100

  assert Bits( 4 ) == Bits( 4, 0 )
  assert Bits( 4 ).uint() == 0

  with pytest.raises( ValueError ):
    a = Bits(4,17)

def test_construct_from_bits():

  assert Bits( 4, Bits(4, -2) ).uint() == 0b1110
  assert Bits( 4, Bits(4, -4) ).uint() == 0b1100

  with pytest.raises( ValueError ):
    a = Bits( 4, Bits(3, -4) )
  with pytest.raises( ValueError ):
    a = Bits( 4, Bits(5, -4) )

  a = Bits( 8, 5 )
  assert a                                  == 0x05
  assert Bits( 16, a.uint() ).uint()        == 0x0005
  assert Bits( 16, (~a + 1).uint() ).uint() == 0x00FB
  b = Bits( 32, 5 )
  assert b                         == 0x00000005
  assert Bits( 32, ~b + 1 ).uint() == 0xFFFFFFFB
  c = Bits( 32, 0 )
  assert c                         == 0x00000000
  assert Bits( 32, ~c )            == 0xFFFFFFFF
  assert Bits( 32, ~c + 1 )        == 0x00000000
  d = Bits( 4, -1 )
  assert Bits( 8, d.uint() )              == 0x0F

  assert Bits( Bits(4,4), 1 ).uint() == 1

def test_str():

  assert Bits(  4,        0x2 ).__str__() == "2"
  assert Bits(  8,       0x1f ).__str__() == "1f"
  assert Bits( 32, 0x0000beef ).__str__() == "0000beef"
  with pytest.raises( ValueError ):
    assert Bits(  4, Bits(32,2) ).__str__() == "2"

def test_index_array():

  data = range( 2**4 )

  # Indexing into an array
  x = Bits( 4, 3  )
  assert data[ x ] == 3

  # Note, this converts -2 to unsigned, so 14!
  y = Bits( 4, -2 )
  assert data[ y ] == 14

  # Larger bitwidths work as long as the list is big enough
  a = Bits( 8, 4  )
  assert data[ a ] == 4

  # If not, regular indexing error
  b = Bits( 8, 20 )
  with pytest.raises( IndexError ):
    data[ b ]

  # Same with negative that become out of range when converted to unsigned
  c = Bits( 8, -1 )
  with pytest.raises( IndexError ):
    data[ c ]

def test_index_bits():

  data = Bits( 8, 0b11001010 )

  # Indexing into a bits
  x = Bits( 4, 3  )
  assert data[ x ] == 1

  # Note, this converts -8 to unsigned, so 8! Out of range!
  y = Bits( 4, -8 )
  with pytest.raises( IndexError ):
    data[ y ]

  # Larger bitwidths work as long as the list is big enough
  a = Bits( 8, 4  )
  assert data[ a ] == 0

  # If not, regular indexing error
  b = Bits( 8, 20 )
  with pytest.raises( IndexError ):
    data[ b ]

  # Same with negative that become out of range when converted to unsigned
  c = Bits( 8, -1 )
  with pytest.raises( IndexError ):
    data[ c ]

def test_slice_bits():

  data = Bits( 8, 0b1101 )

  # Indexing into a bits
  x = Bits( 4, 2  )
  assert data[ : ]   == 0b1101
  assert data[x: ]   == 0b11
  assert data[ :x]   == 0b01
  with pytest.raises( IndexError ):
    assert data[x:x] == 0b1

def test_clone():
  a = Bits(4,3)
  b = a.clone()
  assert a is not b
  assert a == b

def test_deepcopy():
  a = Bits(4,3)
  b = deepcopy( a )
  assert a is not b
  assert a == b

def test_ilshift():

  a = Bits(8,12)
  with pytest.raises( ValueError ):
    a <<= 256
  a <<= 2
  a <<= Bits(8,1)
  with pytest.raises( ValueError ):
    a <<= Bits(7,1)
  with pytest.raises( ValueError ):
    a <<= Bits(9,1)
  a._flip()
  assert a == 1

def test_imatmul():

  a = Bits(8,12)

  with pytest.raises( TypeError ):
    a @ 1

  with pytest.raises( ValueError ):
    a @= 256
  a @= 2
  a @= Bits(8,1)
  assert a == 1
  with pytest.raises( ValueError ):
    a @= Bits(7,1)
  with pytest.raises( ValueError ):
    a @= Bits(9,1)

def test_hash():
  a = Bits(4,12)
  assert hash(a) == hash( (4,12) )

def test_repr():
  assert repr( Bits(15,35) ) == 'Bits15(0x0023)'
  assert repr( Bits(16,35) ) == 'Bits16(0x0023)'
  assert repr( Bits(17,35) ) == 'Bits17(0x00023)'

def test_bin_oct_hex():
  assert Bits(15,35).bin() == "0b000000000100011"
  assert Bits(15,35).oct() == "0o00043"
  assert Bits(15,35).hex() == "0x0023"



from .import_.VerilogVerilatorImportPass import VerilogVerilatorImportPass
from .tbgen.VerilogTBGenPass import VerilogTBGenPass
from .translation.VerilogTranslationPass import VerilogTranslationPass
from .VerilogPlaceholder import VerilogPlaceholder
from .VerilogPlaceholderPass import VerilogPlaceholderPass
from .VerilogTranslationImportPass import VerilogTranslationImportPass



template = \
'''//========================================================================
// V{component_name}_v.cpp
//========================================================================
// This file provides a template for the C wrapper used in the import
// pass. The wrapper exposes a C interface to CFFI so that a
// Verilator-generated C++ model can be driven from Python.
// This templated is based on PyMTL v2.

#include "obj_dir_{component_name}/V{component_name}.h"
#include "stdio.h"
#include "stdint.h"
#include "verilated.h"
#include "verilated_vcd_c.h"

// set to true if the model has clk signal
#define HAS_CLK {has_clk}

// set to true when VCD tracing is enabled in Verilator
#define DUMP_VCD {dump_vcd}

// set to true to enable on-demand VCD dumping
#define ON_DEMAND_DUMP_VCD {on_demand_dump_vcd}

// top level port to be used in on-demand VCD dumping; only dump vars when
// that port has a non-zero value.
#define ON_DEMAND_VCD_ENABLE {on_demand_vcd_enable}

// set to true when Verilog module has line tracing
#define VLINETRACE {external_trace}

#if VLINETRACE
#include "obj_dir_{component_name}/V{component_name}__Syms.h"
#include "svdpi.h"
#endif

//------------------------------------------------------------------------
// CFFI Interface
//------------------------------------------------------------------------
// simulation methods and model interface ports exposed to CFFI

extern "C" {{
  typedef struct {{

    // Exposed port interface
{port_defs}

    // Verilator model
    void * model;

    // VCD state
    int _vcd_en;

    // VCD tracing helpers
    #if DUMP_VCD
    void *        tfp;
    unsigned int  trace_time;
    #endif

  }} V{component_name}_t;

  // Exposed methods
  V{component_name}_t * create_model( const char * );
  void destroy_model( V{component_name}_t *);
  void comb_eval( V{component_name}_t * );
  void seq_eval( V{component_name}_t * );
  void assert_en( bool en );

  #if VLINETRACE
  void trace( V{component_name}_t *, char * );
  #endif

}}

//------------------------------------------------------------------------
// sc_time_stamp
//------------------------------------------------------------------------
// Must be defined so the simulator knows the current time. Called by
// $time in Verilog. See:
// http://www.veripool.org/projects/verilator/wiki/Faq

vluint64_t g_main_time = 0;

double sc_time_stamp()
{{

  return g_main_time;

}}

//------------------------------------------------------------------------
// create_model()
//------------------------------------------------------------------------
// Construct a new verilator simulation, initialize interface signals
// exposed via CFFI, and setup VCD tracing if enabled.

V{component_name}_t * create_model( const char *vcd_filename ) {{

  V{component_name}_t * m;
  V{component_name}   * model;

  Verilated::randReset( {verilator_xinit_value} );
  Verilated::randSeed( {verilator_xinit_seed} );

  m     = (V{component_name}_t *) malloc( sizeof(V{component_name}_t) );
  model = new V{component_name}();

  m->model = (void *) model;

  // Enable tracing. We have added a feature where if the vcd_filename is
  // "" then we don't do any VCD dumping even if DUMP_VCD is true.

  m->_vcd_en = 0;
  #if DUMP_VCD
  if ( strlen( vcd_filename ) != 0 ) {{
    m->_vcd_en = 1;
    Verilated::traceEverOn( true );
    VerilatedVcdC * tfp = new VerilatedVcdC();

    model->trace( tfp, 99 );
    tfp->spTrace()->set_time_resolution( "{vcd_timescale}" );
    tfp->open( vcd_filename );

    m->tfp        = (void *) tfp;
    m->trace_time = 0;
  }}
  #endif

  // initialize exposed model interface pointers
{port_inits}

  return m;

}}

//------------------------------------------------------------------------
// destroy_model()
//------------------------------------------------------------------------
// Finalize the Verilator simulation, close files, call destructors.

void destroy_model( V{component_name}_t * m ) {{

  #if VM_COVERAGE
    VerilatedCov::write( "coverage.dat" );
  #endif

  V{component_name} * model = (V{component_name} *) m->model;

  // finalize verilator simulation
  model->final();

  #if DUMP_VCD
  if ( m->_vcd_en ) {{
    // printf("DESTROYING %d\\n", m->trace_time);
    VerilatedVcdC * tfp = (VerilatedVcdC *) m->tfp;
    tfp->close();
  }}
  #endif

  delete model;

}}

//------------------------------------------------------------------------
// comb_eval()
//------------------------------------------------------------------------
// Simulate one time-step in the Verilated model.

void comb_eval( V{component_name}_t * m ) {{

  V{component_name} * model = (V{component_name} *) m->model;

  // evaluate one time step
  model->eval();

  // Shunning: calling dump multiple times leads to unsuppressable warning
  //           under verilator 4.036
  // #if DUMP_VCD
  // if ( m->_vcd_en ) {{
  //   // dump current signal values
  //   VerilatedVcdC * tfp = (VerilatedVcdC *) m->tfp;
  //   tfp->dump( m->trace_time );
  //   tfp->flush();
  // }}
  // #endif

}}

//------------------------------------------------------------------------
// seq_eval()
//------------------------------------------------------------------------
// Simulate the positive clock edge in the Verilated model.

void seq_eval( V{component_name}_t * m ) {{

  V{component_name} * model = (V{component_name} *) m->model;

  // evaluate one time cycle

  #if HAS_CLK
  model->clk = 0;
  #endif

  model->eval();

  #if DUMP_VCD
  if ( m->_vcd_en && (ON_DEMAND_VCD_ENABLE || !ON_DEMAND_DUMP_VCD) ) {{

    // update simulation time only on clock toggle
    m->trace_time += {half_cycle_time};
    g_main_time   += {half_cycle_time};

    // dump current signal values
    VerilatedVcdC * tfp = (VerilatedVcdC *) m->tfp;
    tfp->dump( m->trace_time );
    tfp->flush();

  }}
  #endif

  #if HAS_CLK
  model->clk = 1;
  #endif

  model->eval();

  #if DUMP_VCD
  if ( m->_vcd_en && (ON_DEMAND_VCD_ENABLE || !ON_DEMAND_DUMP_VCD) ) {{

    // update simulation time only on clock toggle
    m->trace_time += {half_cycle_time};
    g_main_time += {half_cycle_time};

    // dump current signal values
    VerilatedVcdC * tfp = (VerilatedVcdC *) m->tfp;
    tfp->dump( m->trace_time );
    tfp->flush();

  }}
  #endif
}}

//------------------------------------------------------------------------
// assert_en()
//------------------------------------------------------------------------
// Enable or disable assertions controlled by --assert

void assert_en( bool en ) {{

  Verilated::assertOn(en);

}}

//------------------------------------------------------------------------
// trace()
//------------------------------------------------------------------------
// Note that we assume a trace string buffer of 512 characters. This is
// assumed in a couple of places, including the Python wrapper template
// and the Verilog vc/trace.v code. So if we change it, we need to change
// it everywhere.

#if VLINETRACE
void trace( V{component_name}_t * m, char* str ) {{

  V{component_name} * model = (V{component_name} *) m->model;

  const int nchars = 512;
  const int nwords = nchars/4;

  uint32_t words[nwords];
  words[0] = nchars-1;

  // Setup scope for accessing the line tracing function throug DPI.
  // Note, I tried using just this:
  //
  //  svSetScope( svGetScopeFromName("TOP.v.verilog_module") );
  //
  // but it did not seem to work. We would see correct line tracing for
  // the first test case but not any of the remaining tests cases. After
  // digging around a bit, it seemed like the line_trace task was still
  // associated with the very first model as opposed to the newly
  // instantiated models. Directly setting the scope seemed to fix
  // this issue.
  //
  // Note that this issue implies that the mysterious extra .v is no
  // longer present:
  //
  //  https://www.veripool.org/issues/1050-Verilator-Extra-v-in-full-signal-pathnames
  //
  // So now we need to explicitly use the model name instead of
  // Vscope_v__verilog_module.

  // PP: also note that since we add a wrapper around the external Verilog
  // module, the scope we are trying to set is actually the _wrapped_ module
  // which is called `v`.

  svSetScope( &model->__VlSymsp->__Vscope_{vl_component_name}__v );
  model->line_trace( words );

  // Note that the way the line tracing works, the line tracing function
  // will store how the last character used in the line trace in the
  // first element of the word array. The line tracing functions store
  // the line trace starting from the most-signicant character due to the
  // way that Verilog handles strings.

  int nchar_last  = words[0];
  int nchars_used = ( nchars - 1 - nchar_last );

  // We subtract since one of the words (i.e., 4 characters) is for
  // storing the nchars_used.

  assert ( nchars_used < (nchars - 4) );

  // Now we need to iterate from the most-significant character to the
  // last character written by the line tracing functions and copy these
  // characters into the given character array. So we are kind of
  // flipping the order of the characters due to the different between
  // how Verilog and C handle strings.

  int j = 0;
  for ( int i = nchars-1; i > nchar_last; i-- ) {{
    char c = static_cast<char>( words[i/4] >> (8*(i%4)) );
    str[j] = c;
    j++;
  }}
  str[j] = '\\0';

}}
#endif
'''



#=========================================================================
# StructuralRTLIRSignalExpr.py
#=========================================================================
"""RTLIR signal expression class definitions and generation method."""
from pymtl3 import dsl
from pymtl3.datatypes import Bits, is_bitstruct_inst
from pymtl3.passes.rtlir.errors import RTLIRConversionError
from pymtl3.passes.rtlir.rtype import RTLIRDataType as rdt
from pymtl3.passes.rtlir.rtype import RTLIRType as rt

from .StructuralRTLIRGenL0Pass import StructuralRTLIRGenL0Pass


class BaseSignalExpr:
  """Base abstract class of RTLIR signal expressions."""
  def __init__( s, rtype ):
    assert isinstance( rtype, rt.BaseRTLIRType ), f"non-RTLIR type {rtype} encountered!"
    s.rtype = rtype

  def get_rtype( s ):
    return s.rtype

  def __eq__( s, other ):
    return type(s) is type(other) and s.rtype == other.rtype

  def __hash__( s ):
    return hash((type(s), s.rtype))

#-------------------------------------------------------------------------
# Internal signal expression classes
#-------------------------------------------------------------------------

class _Index( BaseSignalExpr ):
  """Base IR class for all index signal operations.

  This class implements a series of static methods that is called to tell
  if the current index signal operation is indexing on a signal of packed
  data type array or a bit selection.
  """
  def __init__( s, index_base, index, rtype ):
    super().__init__( rtype )
    s.index = index
    s.base = index_base

  def __eq__( s, other ):
    return type(s) is type(other) and s.rtype == other.rtype and\
           s.index == other.index and s.base == other.base

  def __hash__( s ):
    return hash((type(s), s.rtype, s.index, s.base))

  def get_index( s ):
    return s.index

  def get_base( s ):
    return s.base

class _UnpackedIndex( _Index ):
  """Base IR class for all index signal operations on unpacked arrays.

  Unpacked indexing operations include indexing on a port array, a wire
  array, a const array, an interface array, and a component array. This class
  implements a series of static methods to check if the given signal operation
  is one of them.
  """
  def __init__( s, index_base, index ):
    base_rtype = index_base.get_rtype()
    rtype = base_rtype.get_next_dim_type()
    super().__init__( index_base, index, rtype )

class _Slice( BaseSignalExpr ):
  """Base IR class for all slicing signal operations.

  This class implements a single static method which checks if the given
  signal operation is slicing a vector signal (a.k.a. a part selection).
  """
  def __init__( s, slice_base, start, stop ):
    base_rtype = slice_base.get_rtype()
    dtype = base_rtype.get_dtype()
    if isinstance( base_rtype, rt.Port ):
      rtype = rt.Port( base_rtype.get_direction(), rdt.Vector( stop-start ) )
    elif isinstance( base_rtype, rt.Wire ):
      rtype = rt.Wire( rdt.Vector( stop-start ) )
    else:
      assert False, f"unrecognized signal type {base_rtype} for slicing"
    super().__init__( rtype )
    s.base = slice_base
    s.slice = ( start, stop )

  def __eq__( s, other ):
    return type(s) is type(other) and s.rtype == other.rtype and \
           s.slice == other.slice and s.base == other.base

  def __hash__( s ):
    return hash((type(s), s.rtype, s.slice, s.base))

  def get_slice( s ):
    return s.slice

  def get_base( s ):
    return s.base

class _Attribute( BaseSignalExpr ):
  """Base IR class for all attribute accessing signal operations.

  This class implements a series of static methods to check if the given signal
  operation is accessing attribute of the current component, a subcompoennt,
  an interface, or a field in a struct signal.
  """
  def __init__( s, attr_base, attr, rtype ):
    super().__init__( rtype )
    s.attr = attr
    s.base = attr_base

  def __eq__( s, other ):
    return type(s) is type(other) and s.rtype == other.rtype and \
           s.attr == other.attr and \
           s.base == other.base

  def __hash__( s ):
    return hash((type(s), s.rtype, s.attr, s.base))

  def get_base( s ):
    return s.base

  def get_attr( s ):
    return s.attr

#-------------------------------------------------------------------------
# Actual IR signal operation classes
#-------------------------------------------------------------------------

class ConstInstance( BaseSignalExpr ):
  """IR class for a constant instance.

  This class is for constants that appear in signal operations and is not
  an attribute of a component.
  """
  def __init__( s, obj, value ):
    super().__init__(rt.Const(rdt.get_rtlir_dtype( obj )))
    s.value = value

  def __eq__( s, other ):
    return isinstance(other, ConstInstance) and s.rtype == other.rtype and \
           s.value == other.value

  def __hash__( s ):
    return hash((type(s), s.rtype, s.value))

  def get_value( s ):
    return s.value

class CurComp( BaseSignalExpr ):
  """IR class for the current component.

  This class is the base of most signal expressions because most signal
  expressions begin by accessing an attribute of the current component.
  This class implements a static method which checks if the given name
  is the same as the current component's name.
  """
  def __init__( s, comp, comp_id ):
    super().__init__(comp.get_metadata(StructuralRTLIRGenL0Pass.rtlir_type))
    s.comp_id = comp_id

  def __eq__( s, other ):
    return isinstance(other, CurComp) and s.rtype == other.rtype and \
           s.comp_id == other.comp_id

  def __hash__( s ):
    return hash((type(s), s.rtype, s.comp_id))

  def get_component_id( s ):
    return s.comp_id

class PortIndex( _UnpackedIndex ):
  """IR class for indexing on a port array ."""

class WireIndex( _UnpackedIndex ):
  """IR class for indexing on a wire array ."""

class ConstIndex( _UnpackedIndex ):
  """IR class for indexing on a const array ."""

class InterfaceViewIndex( _UnpackedIndex ):
  """IR class for indexing on an interface array ."""

class ComponentIndex( _UnpackedIndex ):
  """IR class for indexing on a component array ."""

class PackedIndex( _Index ):
  """IR class for indexing on a signal of packed data type."""
  def __init__( s, index_base, index ):
    base_rtype = index_base.get_rtype()
    dtype = base_rtype.get_dtype()
    if isinstance( base_rtype, rt.Port ):
      rtype = rt.Port( base_rtype.get_direction(),
                    dtype.get_next_dim_type() )
    elif isinstance( base_rtype, rt.Wire ):
      rtype = rt.Wire( dtype.get_next_dim_type() )
    else:
      assert False, f"unrecognized signal type {base_rtype} for indexing"
    super().__init__( index_base, index, rtype )

class BitSelection( _Index ):
  """IR class for selecting a bit of a vector signal."""
  def __init__( s, index_base, index ):
    base_rtype = index_base.get_rtype()
    dtype = base_rtype.get_dtype()
    if isinstance( base_rtype, rt.Port ):
      rtype = rt.Port( base_rtype.get_direction(), rdt.Vector( 1 ) )
    elif isinstance( base_rtype, rt.Wire ):
      rtype = rt.Wire( rdt.Vector( 1 ) )
    else:
      assert False, f"unrecognized signal type {base_rtype} for indexing"
    super().__init__( index_base, index, rtype )

class PartSelection( _Slice ):
  """IR class for selecting one or more bits of a vector signal."""

class CurCompAttr( _Attribute ):
  """IR class for accessing the attribute of the current component."""
  def __init__( s, attr_base, attr ):
    rtype = attr_base.get_rtype().get_property( attr )
    super().__init__( attr_base, attr, rtype )

class SubCompAttr( _Attribute ):
  """IR class for accessing the attribute of a sub-component."""
  def __init__( s, attr_base, attr ):
    rtype = attr_base.get_rtype().get_property( attr )
    super().__init__( attr_base, attr, rtype )

class InterfaceAttr( _Attribute ):
  """IR class for accessing the attribute of an interface."""
  def __init__( s, attr_base, attr ):
    rtype = attr_base.get_rtype().get_property( attr )
    super().__init__( attr_base, attr, rtype )

class StructAttr( _Attribute ):
  """IR class for accessing the attribute of a struct signal."""
  def __init__( s, attr_base, attr ):
    base_rtype = attr_base.get_rtype()
    dtype = base_rtype.get_dtype()
    if isinstance( base_rtype, rt.Port ):
      rtype = rt.Port( base_rtype.get_direction(), dtype.get_property( attr ) )
    elif isinstance( base_rtype, rt.Wire ):
      rtype = rt.Wire( dtype.get_property( attr ) )
    else:
      assert False, f"unrecognized signal type {base_rtype} for field selection"
    super().__init__( attr_base, attr, rtype )

#-------------------------------------------------------------------------
# IR object generation methods for different types
#-------------------------------------------------------------------------

def construct_attr( attr_base, attr ):
  """Return CurCompAttr/SubCompAttr/InterfaceAttr/StructAttr constructor
     if the given attribute belongs to the current component."""
  base_rtype = attr_base.get_rtype()

  if isinstance(attr_base, CurComp):
    if base_rtype.has_property(attr):
      return CurCompAttr( attr_base, attr )
  if isinstance(base_rtype, rt.Component):
    if base_rtype.has_property(attr):
      return SubCompAttr( attr_base, attr )
  if isinstance(base_rtype, rt.InterfaceView):
    if base_rtype.has_property(attr):
      return InterfaceAttr( attr_base, attr )
  if isinstance(base_rtype, rt.Signal):
    dtype = base_rtype.get_dtype()
    if isinstance(dtype, rdt.Struct) and dtype.has_property(attr):
      return StructAttr( attr_base, attr )
  raise AssertionError(f"internal error: no available expression nodes for {attr_base}!")

def construct_index( index_base, index ):
  """Return PackedIndex/BitSelection constructor if the given index base
     is a signal of packed data type."""
  base_rtype = index_base.get_rtype()
  if isinstance(base_rtype, (rt.Port, rt.Wire)):
    dtype = base_rtype.get_dtype()
    if isinstance( dtype, rdt.Vector ):
      return BitSelection( index_base, index )
    elif isinstance( dtype, rdt.PackedArray ):
      return PackedIndex( index_base, index )
  elif isinstance( base_rtype, rt.Array ):
    subtype = base_rtype.get_sub_type()
    if isinstance( subtype, rt.Port ):
      return PortIndex( index_base, index )
    elif isinstance( subtype, rt.Wire ):
      return WireIndex( index_base, index )
    elif isinstance( subtype, rt.Const ):
      return ConstIndex( index_base, index )
    elif isinstance( subtype, rt.InterfaceView ):
      return InterfaceViewIndex( index_base, index )
    elif isinstance( subtype, rt.Component ):
      return ComponentIndex( index_base, index )
  raise AssertionError(f"internal error: no available expression nodes for {index_base}!")

def construct_slice( slice_base, sl ):
  """Return PartSelection constructor if the given index base is a vector signal."""
  base_rtype = slice_base.get_rtype()
  if isinstance(base_rtype, (rt.Port, rt.Wire)) and \
     isinstance(base_rtype.get_dtype(), rdt.Vector):
    return PartSelection( slice_base, sl.start, sl.stop )
  raise AssertionError(f"internal error: no available expression nodes for {slice_base}!")

def construct_base( comp, comp_id ):
  """Return CurComp constructor if comp_id matches the current component's name."""
  return CurComp( comp, comp_id )

#-------------------------------------------------------------------------
# gen_signal_expr
#-------------------------------------------------------------------------

def gen_signal_expr( cur_component, signal ):
  """Return an RTLIR signal expression for the given signal."""

  try:
    if isinstance( signal, dsl.Const ):
      c = signal._dsl.const
      assert isinstance( c, ( int, Bits )) or is_bitstruct_inst( c ), \
                        f'{c} is not an integer/Bits/BitStruct const!'
      return ConstInstance( signal, c )

    assert isinstance( signal, dsl.Signal ), f'{signal} is not supported!'

    # Strip off all the tokens from the signal to the component

    tmp = signal
    stack = []

    if tmp.is_sliced_signal():
      stack.append( (construct_slice, tmp._dsl.slice) )
      tmp = tmp.get_parent_object()

    while tmp is not cur_component:
      if tmp._dsl._my_indices:
        for x in reversed(tmp._dsl._my_indices):
          stack.append( (construct_index, x) )
      stack.append( (construct_attr, tmp._dsl._my_name) )
      tmp = tmp.get_parent_object()

    cur_node = construct_base( tmp, tmp._dsl.my_name )

    for f, token in reversed(stack):
      cur_node = f( cur_node, token )

    return cur_node

  except AssertionError as e:
    msg = '' if e.args[0] is None else e.args[0]
    raise RTLIRConversionError( signal, msg )



#=========================================================================
# VStructuralTranslatorL1.py
#=========================================================================
"""Provide SystemVerilog structural translator implementation."""


from pymtl3.datatypes import Bits
from pymtl3.passes.backends.generic.structural.StructuralTranslatorL1 import (
    StructuralTranslatorL1,
)
from pymtl3.passes.rtlir import RTLIRDataType as rdt
from pymtl3.passes.rtlir import RTLIRType as rt

from ...errors import VerilogPlaceholderError, VerilogReservedKeywordError
from ...util.utility import get_component_unique_name, make_indent, pretty_concat


class VStructuralTranslatorL1( StructuralTranslatorL1 ):

  def check_decl( s, name, msg ):
    if s.is_verilog_reserved( name ):
      raise VerilogReservedKeywordError( name, msg )

  #-----------------------------------------------------------------------
  # Placeholder
  #-----------------------------------------------------------------------

  def rtlir_tr_placeholder_src( s, m ):
    try:
      ph_cfg = m.get_metadata( s._placeholder_pass.placeholder_config )
      if m is s.tr_top:
        # If this placeholder is a top level module, use the wrapper
        # template to support explicit module name.
        if s.tr_cfgs and s.tr_cfgs[m].explicit_module_name:
          module_name = s.tr_cfgs[m].explicit_module_name
        else:
          module_name = ph_cfg.pickled_top_module
          s._mangled_placeholder_top_module_name = module_name

        if module_name == ph_cfg.top_module:
          raise VerilogPlaceholderError(m,
              f"failed to create wrapper for the given object because the same "
              f"name {module_name} is used for both the Verilog top module and "
              f"the wrapper. Please specify a different name through the "
              f"`explicit_module_name` option of TranslationConfigs.")

        # Read the dependency of the placeholder
        nlines     = ph_cfg.pickled_wrapper_nlines
        dependency = '\n'.join( ph_cfg.pickled_source.splitlines()[:-nlines] )

        # Create the placeholder wrapper from the ph_cfg metadata
        wrapper = ph_cfg.pickled_wrapper_template.format(top_module_name=module_name)

        return dependency + wrapper

      else:
        # Otherwise use the pickled placeholder source
        return ph_cfg.pickled_source
    except (AttributeError, OSError):
      # Forgot to apply VerilogPlaceholderPass?
      raise Exception(
                f"error while translating placeholder {m}:\n"
                f"- Did you forget to apply the correct PlaceholderPass(e.g. VerilogPlaceholderPass)?" )

  #-----------------------------------------------------------------------
  # Data types
  #-----------------------------------------------------------------------

  def rtlir_tr_vector_dtype( s, dtype ):
    msb = dtype.get_length() - 1
    return {
      'def'  : '',
      'nbits' : dtype.get_length(),
      'data_type' : 'logic',
      'packed_type' : f'[{msb}:0]',
      'unpacked_type' : '',
      'raw_dtype' : dtype
    }

  def rtlir_tr_unpacked_array_type( s, Type ):
    if Type is None:
      return { 'def' : '', 'unpacked_type' : '', 'n_dim':[] }
    else:
      array_dim = "".join( f"[0:{size-1}]" for size in Type.get_dim_sizes() )
      return {
        'def'  : '',
        'unpacked_type' : array_dim,
        'n_dim' : Type.get_dim_sizes()
      }

  #-----------------------------------------------------------------------
  # Declarations
  #-----------------------------------------------------------------------

  def rtlir_tr_port_decls( s, port_decls ):
    make_indent( port_decls, 1 )
    return ',\n'.join( port_decls )

  def rtlir_tr_port_decl( s, id_, Type, array_type, dtype ):
    _dtype = Type.get_dtype()
    direction = Type.get_direction()
    if direction == 'input':
      direction += ' '
    if array_type:
      template = "Note: port {id_} has data type {_dtype}"
    else:
      n_dim = array_type['n_dim']
      template = "Note: {n_dim} array of ports {id_} has data type {_dtype}"
    s.check_decl( id_, template.format( **locals() ) )
    return pretty_concat( direction, dtype['data_type'], dtype['packed_type'],
              id_, array_type['unpacked_type'] )

  def rtlir_tr_wire_decls( s, wire_decls ):
    make_indent( wire_decls, 1 )
    return '\n'.join( wire_decls )

  def rtlir_tr_wire_decl( s, id_, Type, array_type, dtype ):
    _dtype = Type.get_dtype()
    if array_type:
      template = "Note: wire {id_} has data type {_dtype}"
    else:
      n_dim = array_type['n_dim']
      template = "Note: {n_dim} array of wires {id_} has data type {_dtype}"
    s.check_decl( id_, template.format( **locals() ) )
    return pretty_concat( dtype['data_type'], dtype['packed_type'],
              id_, array_type['unpacked_type'], ';' )

  def rtlir_tr_const_decls( s, const_decls ):
    make_indent( const_decls, 1 )
    return '\n'.join( const_decls )

  def gen_array_param( s, n_dim, dtype, array ):
    if not n_dim:
      if isinstance( dtype, rdt.Vector ):
        return s._literal_number( dtype.get_length(), array )
      else:
        assert False, f'{array} is not an integer or a BitStruct!'
    else:
      ret = []
      for _idx, idx in enumerate( range( n_dim[0] ) ):
        ret.append( s.gen_array_param( n_dim[1:], dtype, array[idx] ) )
      return f"'{{ {', '.join(ret)} }}"

  def rtlir_tr_const_decl( s, id_, Type, array_type, dtype, value ):
    _dtype = Type.get_dtype()
    if array_type:
      template = "Note: constant {id_} has data type {_dtype}"
    else:
      n_dim = array_type['n_dim']
      template = "Note: {n_dim} array of constants {id_} has data type {_dtype}"
    s.check_decl( id_, template.format( **locals() ) )
    _dtype = pretty_concat(dtype['data_type'], dtype['packed_type'], id_, array_type['unpacked_type'])
    _value = s.gen_array_param( array_type['n_dim'], dtype['raw_dtype'], value )

    return f'localparam {_dtype} = {_value};'

  #-----------------------------------------------------------------------
  # Connections
  #-----------------------------------------------------------------------

  def rtlir_tr_connections( s, connections ):
    make_indent( connections, 1 )
    return '\n'.join( connections )

  def rtlir_tr_connection( s, wr_signal, rd_signal ):
    return f'assign {rd_signal} = {wr_signal};'

  #-----------------------------------------------------------------------
  # Signal operations
  #-----------------------------------------------------------------------

  def rtlir_tr_bit_selection( s, base_signal, index, status ):
    # Bit selection
    return s._rtlir_tr_process_unpacked(
              f'{base_signal}[{index}]',
              f'{base_signal}{{}}[{index}]',
              status, ('status', 'unpacked') )

  def rtlir_tr_part_selection( s, base_signal, start, stop, status ):
    # Part selection
    return s._rtlir_tr_process_unpacked(
              f'{base_signal}[{stop-1}:{start}]',
              f'{base_signal}{{}}[{stop-1}:{start}]',
              status, ('status', 'unpacked') )

  def rtlir_tr_port_array_index( s, base_signal, index, status ):
    return s._rtlir_tr_process_unpacked(
              f'{base_signal}[{index}]',
              f'{base_signal}{{}}[{index}]',
              status, ('status', 'unpacked') )

  def rtlir_tr_wire_array_index( s, base_signal, index, status ):
    return f'{base_signal}[{index}]'

  def rtlir_tr_const_array_index( s, base_signal, index, status ):
    return f'{base_signal}[{index}]'

  def rtlir_tr_current_comp_attr( s, base_signal, attr, status ):
    return f'{attr}'

  def rtlir_tr_current_comp( s, comp_id, comp_rtype, status ):
    return ''

  def _rtlir_tr_process_unpacked( s, signal, signal_tplt, status, enable ):
    if (status in ('reader', 'writer') and 'status' in enable) or \
       (s._rtlir_tr_unpacked_q and 'unpacked' in enable):
      ret = signal_tplt.format(''.join(
                [f'[{i}]' for i in list(s._rtlir_tr_unpacked_q)]))
      s._rtlir_tr_unpacked_q.clear()
      return ret
    else:
      return signal

  #-----------------------------------------------------------------------
  # Miscs
  #-----------------------------------------------------------------------

  def rtlir_tr_var_id( s, var_id ):
    return var_id.replace( '[', '__' ).replace( ']', '' )

  def _literal_number( s, nbits, value ):
    return f"{nbits}'d{int(value)}"

  def rtlir_tr_literal_number( s, nbits, value ):
    return s._literal_number( nbits, value )

  def rtlir_tr_component_unique_name( s, c_rtype ):
    return get_component_unique_name( c_rtype )



from ..verilog.VerilogPlaceholder import VerilogPlaceholder as YosysPlaceholder
from ..verilog.VerilogPlaceholderPass import (
    VerilogPlaceholderPass as YosysPlaceholderPass,
)
from .import_.YosysVerilatorImportPass import YosysVerilatorImportPass
from .translation.YosysTranslationPass import YosysTranslationPass
from .YosysTranslationImportPass import YosysTranslationImportPass



from .cl_queues import BypassQueueCL, NormalQueueCL, PipeQueueCL
from .enq_deq_ifcs import DeqIfcRTL, EnqIfcRTL
from .queues import BypassQueueRTL, NormalQueueRTL, PipeQueueRTL

#  from .enrdy_queues import BypassQueue1RTL, NormalQueue1RTL, PipeQueue1RTL
#  from .valrdy_queues import (
    #  BypassQueue1RTL,
    #  NormalQueue1RTL,
    #  NormalQueueRTL,
    #  PipeQueue1RTL,
#  )



"""
========================================================================
MagicMemoryCL
========================================================================
A behavioral magic memory which is parameterized based on the number of
memory request/response ports. This version is a little different from
the one in pclib because we actually use the memory messages correctly
in the interface.

Author : Shunning Jiang
Date   : Feb 6, 2020
"""
from random import Random
from collections import deque

from pymtl3 import *
from pymtl3.extra import clone_deepcopy
from pymtl3.stdlib.delays import DelayPipeDeqCL, DelayPipeSendCL, StallCL

from pymtl3.stdlib.mem.MagicMemoryFL import MagicMemoryFL
from pymtl3.stdlib.mem.MemMsg import MemMsgType, mk_mem_msg
from .ifcs import MinionIfcRTL, RecvIfcRTL, SendIfcRTL
from .queues import NormalQueueRTL, PipeQueueRTL

# BRGTC2 custom MemMsg modified for RISC-V 32

#- - NOTE  - - - NOTE  - - - NOTE  - - - NOTE  - - - NOTE  - - - NOTE  - -
#-------------------------------------------------------------------------
# BRGTC2
#-------------------------------------------------------------------------
# The AMO implementations (and MemMsg) has been updated to match RISC-V.
#
# There is also a small fix to the AMO ops to handle signed ops. The AMO
# operations act on the bitwidth of the processor architecture, so the
# read_data from the TestMemory used with AMOs cannot just be the memory
# request message size (e.g., 128b):
#
#         read_data = Bits( s.data_nbits )
#
# It must instead be the number of bytes matching the bitwidth in the
# processor (e.g., 32b):
#
#         read_data = Bits( nbytes*8 )
#
# Otherwise for example we would be reading 128b from the memory and
# comparing that to the 32b value from the request message.
#
#-------------------------------------------------------------------------
#- - NOTE  - - - NOTE  - - - NOTE  - - - NOTE  - - - NOTE  - - - NOTE  - -

class RandomStall( Component ):
  def construct( s, Type, stall_prob=0, stall_seed=0xdeadbeef ):

    s.recv = RecvIfcRTL( Type )
    s.send = SendIfcRTL( Type )

    s.recv.msg //= s.send.msg

    stall_rgen = Random( stall_seed )

    s.rand_value = 0
    s.stall_prob = stall_prob

    @update_ff
    def up_rand():
      s.rand_value = stall_rgen.random()

    @update
    def up_stall_rdy():
      s.recv.rdy @= s.send.rdy & (s.rand_value > stall_prob)
    @update
    def up_stall_val():
      s.send.val @= s.recv.val & (s.rand_value > stall_prob)

  def line_trace( s ):
    return "[ ]" if s.rand_value > s.stall_prob else "[#]"

# Cuts both val and rdy paths
class InelasticDelayPipe( Component ):

  def construct( s, Type, delay=1 ):

    s.recv = RecvIfcRTL( Type )
    s.send = SendIfcRTL( Type )

    assert delay >= 1
    s.delay = delay

    # FIFO behavior
    s.delay_pipe = deque( [None]*(delay+1) )

    @update_ff
    def up_delay():

      if s.recv.rdy & s.recv.val:
        s.delay_pipe[0] = clone_deepcopy( s.recv.msg )

      # We remove the sent message from the pipe first and then
      # clone the recv message. This allows us to use one entry for one
      # delay

      if s.send.val:
        if s.send.rdy:
          s.delay_pipe[-1] = None
          s.delay_pipe.rotate()
      else: # this cycle invalid send means pipe[-1] is None
        s.delay_pipe.rotate()

      if s.delay_pipe[-1] is None:
        s.send.val <<= 0
      else:
        s.send.val <<= 1
        s.send.msg <<= s.delay_pipe[-1]

      s.recv.rdy <<= s.delay_pipe[0] is None

  def line_trace( s ):
    return f"[{''.join([ ' ' if x is None else '*' for x in s.delay_pipe])}]"

class MagicMemoryRTL( Component ):

  # Magical methods

  def read_mem( s, addr, size ):
    return s.mem.read_mem( addr, size )

  def write_mem( s, addr, data ):
    return s.mem.write_mem( addr, data )

  # Actual stuff
  def construct( s, nports=1, mem_ifc_dtypes=[mk_mem_msg(8,32,32)],
                    stall_prob=0, extra_latency=0, mem_nbytes=2**20 ):

    # Local constants

    assert len(mem_ifc_dtypes) == nports
    req_classes  = [ x for (x,y) in mem_ifc_dtypes ]
    resp_classes = [ y for (x,y) in mem_ifc_dtypes ]

    s.mem = MagicMemoryFL( mem_nbytes )

    # Interface

    s.ifc = [ MinionIfcRTL( req_classes[i], resp_classes[i] ) for i in range(nports) ]


    # stall and delays
    s.req_stalls = [ RandomStall( req_classes[i], stall_prob, i ) for i in range(nports) ]
    # s.req_qs     = [ NormalQueueRTL( req_classes[i], 2 ) for i in range(nports) ]
    s.resp_qs    = [ InelasticDelayPipe( resp_classes[i], extra_latency+1 ) for i in range(nports) ]

    for i in range(nports):
      s.req_stalls[i].recv //= s.ifc[i].req
      # s.req_stalls[i].send //= s.req_qs[i].recv
      s.resp_qs[i].send    //= s.ifc[i].resp

      s.req_stalls[i].send.rdy //= s.resp_qs[i].recv.rdy
      s.req_stalls[i].send.val //= s.resp_qs[i].recv.val

    @update_once
    def up_mem():

      for i in range(nports):

        if s.req_stalls[i].send.val:

          # Dequeue memory request message

          req = s.req_stalls[i].send.msg
          len_ = int(req.len)
          if len_ == 0: len_ = req_classes[i].data_nbits >> 3

          if   req.type_ == MemMsgType.READ:
            resp = resp_classes[i]( req.type_, req.opaque, 0, req.len,
                                    zext( s.mem.read( req.addr, len_ ), req_classes[i].data_nbits ) )

          elif  req.type_ == MemMsgType.WRITE:
            s.mem.write( req.addr, len_, req.data[0:len_<<3] )
            # FIXME do we really set len=0 in response when doing subword wr?
            # resp = resp_classes[i]( req.type_, req.opaque, 0, req.len, 0 )
            resp = resp_classes[i]( req.type_, req.opaque, 0, 0, 0 )

          elif  req.type_ == MemMsgType.AMO_ADD   or \
                req.type_ == MemMsgType.AMO_AND   or \
                req.type_ == MemMsgType.AMO_MAX   or \
                req.type_ == MemMsgType.AMO_MAXU  or \
                req.type_ == MemMsgType.AMO_MIN   or \
                req.type_ == MemMsgType.AMO_MINU  or \
                req.type_ == MemMsgType.AMO_OR    or \
                req.type_ == MemMsgType.AMO_SWAP  or \
                req.type_ == MemMsgType.AMO_XOR:
            resp = resp_classes[i]( req.type_, req.opaque, 0, req.len,
               s.mem.amo( req.type_, req.addr, len_, req.data ) )

          # Invalid type
          else:
            assert False

          s.resp_qs[i].recv.msg @= resp

  #-----------------------------------------------------------------------
  # line_trace
  #-----------------------------------------------------------------------

  def line_trace( s ):
    # print()
    return "|".join( f"{s.req_stalls[i].line_trace()}{s.ifc[i].req}>{s.ifc[i].resp}{s.resp_qs[i].line_trace()}"
                        for i in range(len(s.ifc)) )



#=========================================================================
# YosysBehavioralTranslatorL5_test.py
#=========================================================================
# Author : Peitian Pan
# Date   : June 9, 2019
"""Test the YosysVerilog translator implementation."""

import pytest

from pymtl3.passes.backends.verilog.util.utility import verilog_reserved
from pymtl3.passes.rtlir import BehavioralRTLIRGenPass, BehavioralRTLIRTypeCheckPass

from ....testcases import (
    CaseBehavioralArraySubCompArrayStructIfcComp,
    CaseBits32ArraySubCompAttrUpblkComp,
    CaseBits32SubCompAttrUpblkComp,
)
from ..YosysBehavioralTranslatorL5 import YosysBehavioralRTLIRToVVisitorL5


def run_test( case, m ):
  m.elaborate()
  visitor = YosysBehavioralRTLIRToVVisitorL5(lambda x: x in verilog_reserved)

  m.apply( BehavioralRTLIRGenPass( m ) )
  m.apply( BehavioralRTLIRTypeCheckPass( m ) )
  upblks = m.get_metadata( BehavioralRTLIRGenPass.rtlir_upblks )
  m_all_upblks = m.get_update_blocks()
  assert len(m_all_upblks) == 1

  for blk in m_all_upblks:
    upblk_src = visitor.enter( blk, upblks[blk] )
    upblk_src = "\n".join( upblk_src )
    assert upblk_src + '\n' == case.REF_UPBLK

@pytest.mark.parametrize(
    'case', [
      CaseBits32SubCompAttrUpblkComp,
      CaseBits32ArraySubCompAttrUpblkComp,
      CaseBehavioralArraySubCompArrayStructIfcComp,
    ]
)
def test_yosys_behavioral_L5( case ):
  run_test( case, case.DUT() )



#=========================================================================
# VBehavioralTranslatorL4.py
#=========================================================================
# Author : Peitian Pan
# Date   : March 18, 2019
"""Provide the level 4 SystemVerilog translator implementation."""

from pymtl3.passes.backends.generic.behavioral.BehavioralTranslatorL4 import (
    BehavioralTranslatorL4,
)
from pymtl3.passes.rtlir import RTLIRType as rt

from ...errors import VerilogTranslationError
from .VBehavioralTranslatorL3 import (
    BehavioralRTLIRToVVisitorL3,
    VBehavioralTranslatorL3,
)


class VBehavioralTranslatorL4(
    VBehavioralTranslatorL3, BehavioralTranslatorL4):

  def _get_rtlir2v_visitor( s ):
    return BehavioralRTLIRToVVisitorL4

#-------------------------------------------------------------------------
# BehavioralRTLIRToVVisitorL4
#-------------------------------------------------------------------------

class BehavioralRTLIRToVVisitorL4( BehavioralRTLIRToVVisitorL3 ):
  """Visitor that translates RTLIR to SystemVerilog for a single upblk."""

  #-----------------------------------------------------------------------
  # visit_Attribute
  #-----------------------------------------------------------------------

  def visit_Attribute( s, node ):
    """Return the SystemVerilog representation of an attribute.

    Add support for interface attributes in L4. We just name mangle every
    signal in an interface instead of constructing a SystemVerilog interface
    and instantiating new interfaces from it.
    """
    if isinstance( node.value.Type, rt.InterfaceView ):
      value = s.visit( node.value )
      attr = node.attr
      s.check_res( node, attr )
      return s.process_unpacked_q( node,
                f'{value}__{attr}',
                f'{value}__{attr}{{}}' )
    else:
      return super().visit_Attribute( node )

  #-----------------------------------------------------------------------
  # visit_Index
  #-----------------------------------------------------------------------

  def visit_Index( s, node ):
    if isinstance( node.value.Type, rt.Array ) and \
        isinstance( node.value.Type.get_sub_type(), rt.InterfaceView ):
      idx = s.visit( node.idx )
      s._unpacked_q.appendleft(idx)
      value = s.visit( node.value )
      return value

    else:
      return super().visit_Index( node )



"""
========================================================================
SimpleTickPass.py
========================================================================
Generate a simple tick function (no Mamba techniques here)

Author : Shunning Jiang
Date   : Dec 26, 2018
"""
from pymtl3.dsl import MethodPort
from pymtl3.dsl.errors import UpblkCyclicError
from pymtl3.passes.BasePass import BasePass
from pymtl3.passes.errors import PassOrderError

# Shunning: We are aware of the problem that there may be multiple places
# that assembles the function for the user to tick the simulator.
# I think LLVM applies passes based on the order they are inserted to the
# pass manager. LLVM also has some local ordering mechanism to figure out
# the dependencies between passes. Currently we haven't got to the point
# where we have enough passes to manage, but we should keep this in mind.

class SimpleTickPass( BasePass ):

  @staticmethod
  def gen_tick_function( schedule ):
    def iterative():
      for blk in schedule:
        blk()
    return iterative



from .autotick.OpenLoopCLPass import OpenLoopCLPass
from .BasePass import BasePass
from .sim.DynamicSchedulePass import DynamicSchedulePass
from .sim.GenDAGPass import GenDAGPass
from .sim.PrepareSimPass import PrepareSimPass
from .sim.SimpleSchedulePass import SimpleSchedulePass
from .sim.SimpleTickPass import SimpleTickPass
from .sim.WrapGreenletPass import WrapGreenletPass
from .tracing.CLLineTracePass import CLLineTracePass
from .tracing.LineTraceParamPass import LineTraceParamPass
from .tracing.PrintTextWavePass import PrintTextWavePass
from .tracing.VcdGenerationPass import VcdGenerationPass


# SimpleSim can be used when the UDG is a DAG
class SimpleSimPass( BasePass ):
  def __call__( s, top ):
    LineTraceParamPass()( top )
    GenDAGPass()( top )
    WrapGreenletPass()( top )
    SimpleSchedulePass()( top )
    CLLineTracePass()( top )
    VcdGenerationPass()( top )
    PrintTextWavePass()( top )

    PrepareSimPass(print_line_trace=False)( top )

class DefaultPassGroup( BasePass ):
  def __init__( s, *, vcdwave=None, textwave=False,
                      linetrace=False, reset_active_high=True ):

    s.vcdwave = vcdwave
    s.textwave = textwave
    s.linetrace = linetrace
    s.reset_active_high = reset_active_high

  def __call__( s, top ):

    if s.vcdwave:
      top.set_metadata( VcdGenerationPass.vcd_file_name, s.vcdwave )

    if s.textwave:
      top.set_metadata( PrintTextWavePass.enable, True )

    LineTraceParamPass()( top )
    GenDAGPass()( top )
    WrapGreenletPass()( top )
    CLLineTracePass()( top )
    DynamicSchedulePass()( top )
    VcdGenerationPass()( top )
    PrintTextWavePass()( top )

    PrepareSimPass(print_line_trace=s.linetrace,
                   reset_active_high=s.reset_active_high)( top )

class AutoTickSimPass( BasePass ):
  def __init__( s, print_line_trace=True ):
    s.print_line_trace = print_line_trace

  def __call__( s, top ):
    top.elaborate()
    GenDAGPass()( top )
    WrapGreenletPass()( top )
    OpenLoopCLPass( s.print_line_trace )( top )
    top.lock_in_simulation()



#=========================================================================
# fast_bytearray_funcs.py
#=========================================================================
#
# Author : Shunning Jiang
# Date   : Feb 25, 2020

try:
  from mamba import read_bytearray_bits
except:
  from pymtl3.datatypes import Bits

  def read_bytearray_bits( arr, addr, nbytes ):
    ret = 0
    begin = int(addr)
    addr  = begin + nbytes - 1

    while addr >= begin:
      ret = (ret << 8) + arr[addr]
      addr -= 1

    return Bits( nbytes << 3, ret )

try:
  from mamba import write_bytearray_bits
except:
  from pymtl3.datatypes import Bits

  def write_bytearray_bits( arr, addr, nbytes, data ):
    addr = int(addr)
    end  = addr + nbytes

    while addr < end:
      arr[addr] = data & 255
      data >>= 8
      addr += 1



"""
========================================================================
enrdy_queues.py
========================================================================
This file contains queues with EnRdy (send/recv) interface.

Author : Shunning Jiang
Date   : Mar 9, 2018
"""

from pymtl3 import *
from pymtl3.stdlib.basic_rtl import Mux, Reg, RegEn, RegRst
from pymtl3.stdlib.ifcs import RecvIfcRTL, SendIfcRTL


class PipeQueue1RTL( Component ):

  def construct( s, Type ):
    s.enq = RecvIfcRTL( Type )
    s.deq = SendIfcRTL( Type )

    s.buffer = m = RegEn( Type )
    m.en  //= s.enq.en
    m.in_ //= s.enq.msg
    m.out //= s.deq.msg

    s.full = Reg( Bits1 )

    @update
    def up_pipeq_use_deq_rdy():
      s.deq.en  @=  s.full.out & s.deq.rdy
      s.enq.rdy @= ~s.full.out | s.deq.rdy

    @update
    def up_pipeq_full():
      s.full.in_ @= s.enq.en | (s.full.out & ~s.deq.rdy )

  def line_trace( s ):
    return s.buffer.line_trace()

class BypassQueue1RTL( Component ):

  def construct( s, Type ):
    s.enq = RecvIfcRTL( Type )
    s.deq = SendIfcRTL( Type )

    s.buffer = RegEn( Type )
    s.buffer.in_ //= s.enq.msg

    s.full = RegRst( Bits1, reset_value = 0 )

    s.byp_mux = m = Mux( Type, 2 )
    m.out    //= s.deq.msg
    m.in_[0] //= s.enq.msg
    m.in_[1] //= s.buffer.out
    m.sel    //= s.full.out # full -- buffer.out, empty -- bypass

    @update
    def up_bypq_set_enq_rdy():
      s.enq.rdy @= ~s.full.out

    @update
    def up_bypq_use_enq_en():
      s.deq.en    @= (s.enq.en | s.full.out) & s.deq.rdy
      s.buffer.en @=  s.enq.en & ~s.deq.en
      s.full.in_  @= (s.enq.en | s.full.out) & ~s.deq.en

  def line_trace( s ):
    return s.buffer.line_trace()

class BypassQueue2RTL( Component ):

  def construct( s, MsgType, queue_size=2 ):
    assert queue_size == 2
    s.enq = RecvIfcRTL( MsgType )
    s.deq = SendIfcRTL( MsgType )
    s.q1 = BypassQueue1RTL( MsgType )
    s.q2 = BypassQueue1RTL( MsgType )

    s.enq    //= s.q1.enq
    s.q1.deq //= s.q2.enq
    s.q2.deq //= s.deq

  def line_trace( s ):
    return "{}({}){}".format( s.enq, s.q1.deq, s.deq )

class NormalQueue1RTL( Component ):

  def construct( s, Type ):
    s.enq = RecvIfcRTL( Type )
    s.deq = SendIfcRTL( Type )

    # Now since enq.en depends on enq.rdy, enq.en == 1 actually means
    # we will enq some stuff
    s.buffer = m = RegEn( Type )
    m.en  //= s.enq.en
    m.in_ //= s.enq.msg
    m.out //= s.deq.msg
    s.full = Reg( Bits1 )

    @update
    def up_normq_set_enq_rdy():
      s.enq.rdy @= ~s.full.out

    @update
    def up_normq_full():
      # When the bufferf is full and deq side is ready, we enable deq
      s.deq.en @= s.full.out & s.deq.rdy

      # not full and valid enq, or no deque and enq, or no deque and already full
      s.full.in_ @= (~s.full.out & s.enq.en) | \
                    (~s.deq.rdy & s.enq.en)  | \
                    (~s.deq.rdy & s.full.out)


  def line_trace( s ):
    return s.buffer.line_trace()



"""
========================================================================
test_utils
========================================================================
Helper functions for running unit tests

Author : Shunning Jiang
  Date : Jan 23, 2020
"""

import collections
import re

from pymtl3 import *
from pymtl3.datatypes import is_bitstruct_class
from pymtl3.passes.backends.verilog import *
from pymtl3.passes.tracing import VcdGenerationPass, PrintTextWavePass

#-------------------------------------------------------------------------
# mk_test_case_table
#-------------------------------------------------------------------------

def mk_test_case_table( raw_test_case_table ):

  # First row in test vectors contains port names

  if isinstance(raw_test_case_table[0],str):
    test_param_names = raw_test_case_table[0].split()
  else:
    test_param_names = raw_test_case_table[0]

  TestCase = collections.namedtuple("TestCase",test_param_names)

  ids = []
  test_cases = []
  for row in raw_test_case_table[1:]:
    ids.append( row[0] )
    test_cases.append( TestCase(*row[1:]) )

  return {
    'ids'      : ids,
    'argnames' : ('test_params'),
    'argvalues' : test_cases,
  }


def finalize_verilator( model ):
  for child in model.get_child_components():
    finalize_verilator( child )
  if hasattr( model, 'finalize' ):
    model.finalize()

def _recursive_set_vl_trace( m, dump_vcd ):
  if ( m.has_metadata( VerilogTranslationImportPass.enable ) and \
       m.get_metadata( VerilogTranslationImportPass.enable ) ) or \
      isinstance( m, VerilogPlaceholder ):
    m.set_metadata( VerilogVerilatorImportPass.vl_trace, True )
    vl_trace_filename = f"top{repr(m)[1:]}".replace('.','_')
    m.set_metadata( VerilogVerilatorImportPass.vl_trace_filename, dump_vcd+"_"+vl_trace_filename )
  else:
    for child in m.get_child_components():
      _recursive_set_vl_trace( child, dump_vcd )

def config_model_with_cmdline_opts( top, cmdline_opts, duts ):

  test_verilog  = cmdline_opts[ 'test_verilog' ]
  dump_textwave = cmdline_opts[ 'dump_textwave' ]
  dump_vcd      = cmdline_opts[ 'dump_vcd'     ]
  dump_vtb      = cmdline_opts[ 'dump_vtb'     ]

  if 'on_demand_vcd_portname' in cmdline_opts:
    on_demand_vcd_portname = cmdline_opts['on_demand_vcd_portname']
  else:
    on_demand_vcd_portname = ""

  # Setup the model

  top.elaborate()

  if dump_vcd:
    _recursive_set_vl_trace( top, dump_vcd )

  if on_demand_vcd_portname:
    assert test_verilog, "--test-verilog has to be present to enable on-demand VCD!"

  if duts:
    dut_objs = []
    for i, dut in enumerate(duts):
      dut_objs.append( eval(f'top.{dut}') )
  else:
    dut_objs = [ top ]

  for dut in dut_objs:
    if test_verilog:
      dut.set_metadata( VerilogTranslationImportPass.enable, True )
      dut.set_metadata( VerilogVerilatorImportPass.vl_xinit, test_verilog )

      if dump_vcd:
        dut.set_metadata( VerilogVerilatorImportPass.vl_trace, True )
        dut.set_metadata( VerilogVerilatorImportPass.vl_trace_filename, dump_vcd )

      if on_demand_vcd_portname:
        dut.set_metadata( VerilogVerilatorImportPass.vl_trace_on_demand, True )
        dut.set_metadata( VerilogVerilatorImportPass.vl_trace_on_demand_portname, on_demand_vcd_portname )

  top.apply( VerilogPlaceholderPass() )
  top = VerilogTranslationImportPass()( top )

  # Access dut object again because those dut_objs can be replaced by
  # import passes
  if duts:
    dut_objs = []
    for i, dut in enumerate(duts):
      dut_objs.append( eval(f'top.{dut}') )
  else:
    dut_objs = [ top ]

  if dump_vtb:
    for dut in dut_objs:
      dut.set_metadata( VerilogTBGenPass.case_name, dump_vtb )

    top.apply( VerilogTBGenPass() )

  # FIXME we currently have to set dump_vcd after applying translationimport pass
  # Need to transfer metadata from the new DUT
  if dump_vcd:
    top.set_metadata( VcdGenerationPass.vcd_file_name, dump_vcd )

  if dump_textwave:
    top.set_metadata( PrintTextWavePass.enable, True )

  return top

#------------------------------------------------------------------------------
# TestVectorSimulator
#------------------------------------------------------------------------------
# Now we have the same test vector simulator as pymtl2
# https://github.com/cornell-brg/pymtl/blob/master/pclib/test/TestVectorSimulator.py

class TestVectorSimulator:

  def __init__( self, model, test_vectors,
                set_inputs_func, verify_outputs_func, wait_cycles = 0 ):

    self.model               = model
    self.set_inputs_func     = set_inputs_func
    self.verify_outputs_func = verify_outputs_func
    self.test_vectors        = test_vectors
    self.wait_cycles         = wait_cycles

  def run_test( self, cmdline_opts=None ):

    cmdline_opts = cmdline_opts or {'dump_textwave' : False,
                                    'dump_vcd'      : False,
                                    'test_verilog'  : False,
                                    'dump_vtb'      : ''}

    # Setup the model
    self.model = config_model_with_cmdline_opts( self.model, cmdline_opts, [] )

    try:
      self.model.apply( DefaultPassGroup(linetrace=True) )
      self.model.sim_reset()

      for test_vector in self.test_vectors:
        # Set inputs
        self.set_inputs_func( self.model, test_vector )

        self.model.sim_eval_combinational()

        try:
          # Verify outputs
          self.verify_outputs_func( self.model, test_vector )
        except Exception as e:
          self.model.print_line_trace()
          raise e

        self.model.sim_tick()

    finally:
      # Dump out textwave at the end of simulation
      if cmdline_opts['dump_textwave']:
          self.model.print_textwave()

      finalize_verilator( self.model )

def run_sim( model, cmdline_opts=None, print_line_trace=True, duts=None ):

  cmdline_opts = cmdline_opts or {'dump_textwave' : False,
                                  'dump_vcd'      : False,
                                  'test_verilog'  : False,
                                  'max_cycles'    : None,
                                  'dump_vtb'      : ''}

  max_cycles = cmdline_opts['max_cycles'] or 10000

  # Setup the model

  model = config_model_with_cmdline_opts( model, cmdline_opts, duts )

  try:
    # Create a simulator
    model.apply( DefaultPassGroup(linetrace=print_line_trace) )
    # Reset model
    model.sim_reset()

    # Run simulation
    while not model.done() and model.sim_cycle_count() < max_cycles:
      model.sim_tick()

    # Force a test failure if we timed out
    assert model.sim_cycle_count() < max_cycles

    # Extra ticks to make VCD easier to read
    model.sim_tick()
    model.sim_tick()
    model.sim_tick()

  finally:
    # Dump out textwave at the end of simulation
    if cmdline_opts['dump_textwave']:
        model.print_textwave()

    finalize_verilator( model )

class RunTestVectorSimError( Exception ):
  pass

def run_test_vector_sim( model, test_vectors, cmdline_opts=None, print_line_trace=True ):
  cmdline_opts = cmdline_opts or {'dump_textwave' : False,
                                  'dump_vcd'      : False,
                                  'test_verilog'  : False,
                                  'dump_vtb'      : ''}

  # First row in test vectors contains port names

  if isinstance(test_vectors[0],str):
    port_names = test_vectors[0].split()
  else:
    port_names = test_vectors[0]

  # Remaining rows contain the actual test vectors

  test_vectors = test_vectors[1:]

  # Setup the model

  model = config_model_with_cmdline_opts( model, cmdline_opts, [] )

  try:
    # Create a simulator
    model.apply( DefaultPassGroup(linetrace=print_line_trace) )
    # Reset model
    model.sim_reset()

    # Run the simulation

    row_num = 0
    in_ids  = []
    out_ids = []
    groups  = [ None ] * len(port_names)
    types   = [ None ] * len(port_names)

    # Preprocess default type
    # Special case for lists of ports
    # NOTE THAT WE ONLY SUPPORT 1D ARRAY and no interface
    for i, port_full_name in enumerate( port_names ):
      if port_full_name[-1] == "*":
        out_ids.append( i )
        port_name = port_full_name[:-1]
      else:
        in_ids.append( i )
        port_name = port_full_name

      if '[' in port_name:
        # Get tokens of the full name
        m = re.match( r'(\w+)\[(\d+)\]', port_name )
        if not m:
          raise Exception(f"Could not parse port name: {port_name}. "
                          f"Currently we don't support interface or high-D array.")

        groups[i] = g = ( True, m.group(1), int(m.group(2)) )

        if not hasattr( model, g[1] ):
          raise RunTestVectorSimError(f"Invalid port name: {g[1]}")

        # Get type of all the ports
        t = type( getattr( model, g[1] )[ int(g[2]) ] )
        types[i] = None if is_bitstruct_class( t ) else t

      else:
        groups[i] = ( False, port_name )

        if not hasattr( model, port_name ):
          raise RunTestVectorSimError(f"Invalid port name: {port_name}")

        t = type( getattr( model, port_name ) )
        types[i] = None if is_bitstruct_class( t ) else t

    # Run simulation

    for row in test_vectors:
      row_num += 1

      # Apply test inputs
      for i in in_ids:
        in_value = row[i]
        if in_value == '?':
          raise RunTestVectorSimError(f"""
Invalid input value in row {row_num} ({row}:
- '?' can only appear in output values (labeled with '*' in the port name specifications).

Please double check the provided values.
""" )
        t = types[i]
        if t: in_value = t( in_value )
        g = groups[i]
        x = getattr( model, g[1] )
        if g[0]:  x[g[2]] @= in_value
        else:     x       @= in_value

      # Evaluate combinational concurrent blocks
      model.sim_eval_combinational()

      # Check test outputs
      for i in out_ids:
        ref_value = row[i]
        if ref_value == '?':  continue

        g = groups[i]
        if g[0]:  out_value = getattr( model, g[1] )[g[2]]
        else:     out_value = getattr( model, g[1] )

        if out_value != ref_value:
          if print_line_trace:
            model.print_line_trace()

          port_name = g[1]
          if g[0]:
            port_name = f"{g[1]}[{g[2]}]"

          error_msg = """
run_test_vector_sim received an incorrect value!
- row number     : {row_number}
- port name      : {port_name}
- expected value : {expected_msg}
- actual value   : {actual_msg}
"""
          raise RunTestVectorSimError( error_msg.format(
            row_number   = row_num,
            port_name    = port_name,
            expected_msg = ref_value,
            actual_msg   = out_value
          ))

      # Tick the simulation
      model.sim_tick()

    # Extra ticks to make VCD easier to read
    model.sim_tick()
    model.sim_tick()
    model.sim_tick()

  finally:
    # Dump out textwave at the end of simulation
    if cmdline_opts['dump_textwave']:
        model.print_textwave()

    finalize_verilator( model )



from pymtl3 import *

# N-input Mux

class Mux( Component ):

  def construct( s, Type, ninputs ):
    assert ninputs > 0
    s.in_ = [ InPort( Type ) for _ in range(ninputs) ]
    s.out = OutPort( Type )
    s.sel = InPort( max(1, clog2(ninputs)) ) # allow 1-input

    @update
    def up_mux():
      s.out @= s.in_[ s.sel ]

# N-output Demux

class Demux( Component ):

  def construct( s, Type, noutputs ):
    assert noutputs > 0
    s.in_ = InPort( Type )
    s.out = [ OutPort( Type ) for _ in range(noutputs) ]
    s.sel = InPort( max(1, clog2(noutputs)) ) # allow 1-input

    default_value = Type()

    @update
    def up_mux():
      for i in range(noutputs):
        s.out[i] @= default_value
      s.out[ s.sel ] @= s.in_

# Rshifter

class RightLogicalShifter( Component ):

  def construct( s, Type, shamt_nbits=0 ):
    if shamt_nbits == 0: shamt_nbits = Type.nbits
    assert shamt_nbits == Type.nbits

    s.in_   = InPort( Type )
    s.shamt = InPort( shamt_nbits )
    s.out   = OutPort( Type )

    @update
    def up_rshifter():
      s.out @= s.in_ >> s.shamt

# Lshifter

class LeftLogicalShifter( Component ):

  def construct( s, Type, shamt_nbits=0 ):
    if shamt_nbits == 0: shamt_nbits = Type.nbits
    assert shamt_nbits == Type.nbits

    s.in_   = InPort( Type )
    s.shamt = InPort( shamt_nbits )
    s.out   = OutPort( Type )
    assert shamt_nbits == Type.nbits

    @update
    def up_lshifter():
      s.out @= s.in_ << s.shamt

# Incrementer

class Incrementer( Component ):

  def construct( s, Type, amount=1 ):
    s.in_ = InPort( Type )
    s.out = OutPort( Type )

    @update
    def up_incrementer():
      s.out @= s.in_ + amount

# Adder

class Adder( Component ):

  def construct( s, Type ):
    s.in0 = InPort( Type )
    s.in1 = InPort( Type )
    s.out = OutPort( Type )

    @update
    def up_adder():
      s.out @= s.in0 + s.in1

# And

class And( Component ):

  def construct( s, Type ):
    s.in0 = InPort( Type )
    s.in1 = InPort( Type )
    s.out = OutPort( Type )

    @update
    def up_and():
      s.out @= s.in0 & s.in1

# Subtractor

class Subtractor( Component ):

  def construct( s, Type ):
    s.in0 = InPort( Type )
    s.in1 = InPort( Type )
    s.out = OutPort( Type )

    @update
    def up_subtractor():
      s.out @= s.in0 - s.in1

# ZeroComparator

class ZeroComparator( Component ):

  def construct( s, Type ):
    s.in_ = InPort( Type )
    s.out = OutPort()

    @update
    def up_zerocomp():
      s.out @= s.in_ == 0

# LeftThanComparator

class LTComparator( Component ):

  def construct( s, Type ):
    s.in0 = InPort( Type )
    s.in1 = InPort( Type )
    s.out = OutPort()

    @update
    def up_ltcomp():
      s.out @= s.in0 < s.in1

# LeftThanOrEqualToComparator

class LEComparator( Component ):

  def construct( s, Type ):
    s.in0 = InPort( Type )
    s.in1 = InPort( Type )
    s.out = OutPort()

    @update
    def up_lecomp():
      s.out @= s.in0 <= s.in1

# EqComparator

class EqComparator( Component ):

  def construct( s, Type ):
    s.in0 = InPort( Type )
    s.in1 = InPort( Type )
    s.out = OutPort()

    @update
    def up_lecomp():
      s.out @= s.in0 == s.in1



#=========================================================================
# TranslationImport_closed_loop_directed_test.py
#=========================================================================
# Author : Peitian Pan
# Date   : Jun 5, 2019
"""Closed-loop test with SystemVerilog translation and import."""

from pymtl3.datatypes import Bits1, mk_bits
from pymtl3.passes.PassGroups import DefaultPassGroup
from pymtl3.passes.rtlir.util.test_utility import do_test
from pymtl3.stdlib.queues.test.enrdy_queues_test import (
    test_bypass_queue as _bypass_queue,
)
from pymtl3.stdlib.queues.test.enrdy_queues_test import (
    test_bypass_queue_stall as _bypass_queue_stall,
)
from pymtl3.stdlib.queues.test.enrdy_queues_test import (
    test_normal_queue as _normal_queue,
)
from pymtl3.stdlib.queues.test.enrdy_queues_test import (
    test_normal_queue_stall as _normal_queue_stall,
)
from pymtl3.stdlib.queues.test.enrdy_queues_test import test_pipe_queue as _pipe_queue
from pymtl3.stdlib.queues.test.enrdy_queues_test import (
    test_pipe_queue_stall as _pipe_queue_stall,
)

from .. import VerilogTranslationImportPass
from ..util.test_utility import closed_loop_component_input_test

#-------------------------------------------------------------------------
# Valrdy queue tests
#-------------------------------------------------------------------------

def run_sim( _th ):
  _th.elaborate()
  _th.q.set_metadata( VerilogTranslationImportPass.enable, True )
  th = VerilogTranslationImportPass()( _th )

  try:
    th.apply( DefaultPassGroup() )
    th.sim_reset()

    while not th.done() and th.sim_cycle_count() < 1000:
      th.sim_tick()

    assert th.sim_cycle_count() < 1000

    th.sim_tick()
    th.sim_tick()
    th.sim_tick()
  finally:
    th.q.finalize()

def _run_queue_test_replace_run_sim( run_sim, test_func ):
  original_run_sim = test_func.__globals__['run_sim']
  test_func.__globals__['run_sim'] = run_sim
  try:
    test_func()
  finally:
    test_func.__globals__['run_sim'] = original_run_sim

def test_normal_queue():
  _run_queue_test_replace_run_sim( run_sim, _normal_queue )

def test_normal_queue_stall():
  _run_queue_test_replace_run_sim( run_sim, _normal_queue_stall )

def test_pipe_queue():
  _run_queue_test_replace_run_sim( run_sim, _pipe_queue )

def test_pipe_queue_stall():
  _run_queue_test_replace_run_sim( run_sim, _pipe_queue_stall )

def test_bypass_queue():
  _run_queue_test_replace_run_sim( run_sim, _bypass_queue )

def test_bypass_queue_stall():
  _run_queue_test_replace_run_sim( run_sim, _bypass_queue_stall )



#=========================================================================
# PlaceholderPass.py
#=========================================================================
# Author : Peitian Pan
# Date   : Jan 27, 2020

from pymtl3 import MetadataKey, Placeholder

from .BasePass import BasePass
from .errors import PlaceholderConfigError


class PlaceholderPass( BasePass ):

  # Placeholder pass input pass data

  #: Enable pickling on the placeholder component.
  #:
  #: Type: ``bool``; input
  #:
  #: Default value: ``False``
  enable             = MetadataKey(bool)

  #: Does the module of external source have clk pin?
  #:
  #: Type: ``bool``; input
  #:
  #: Default value: ``False``
  has_clk            = MetadataKey(bool)

  #: Does the module of external source have reset pin?
  #:
  #: Type: ``bool``; input
  #:
  #: Default value: ``False``
  has_reset          = MetadataKey(bool)

  # Placeholder pass output pass data

  #: An instance of :class:`PlaceholderConfigs` that contains the parsed options.
  #:
  #: Type: ``PlaceholderConfigs``; output
  placeholder_config = MetadataKey()

  def __call__( s, m ):
    """Pickle every ``Placeholder`` in the component hierarchy rooted at ``m``."""
    if isinstance( m, Placeholder ):
      s.visit_placeholder( m )
    for child in m.get_child_components(repr):
      s.__call__( child )

  def visit_placeholder( s, m ):
    s.setup_configs( m )

  def setup_configs( s, m ):
    c = s.__class__
    ph_config = c.get_placeholder_config()( m )
    m.set_metadata( c.placeholder_config, ph_config )

  @staticmethod
  def get_placeholder_config():
    from pymtl3.passes.PlaceholderConfigs import PlaceholderConfigs
    return PlaceholderConfigs



#=========================================================================
# VStructuralTranslatorL1_test.py
#=========================================================================
# Author : Peitian Pan
# Date   : May 29, 2019
"""Test the level 1 SystemVerilog structural translator."""

from collections import deque

import pytest

from pymtl3.passes.backends.verilog.util.test_utility import check_eq
from pymtl3.passes.backends.verilog.util.utility import verilog_reserved

from ....testcases import (
    CaseBitSelOverBitSelComp,
    CaseBitSelOverPartSelComp,
    CaseConnectBitsConstToOutComp,
    CaseConnectBitSelToOutComp,
    CaseConnectConstToOutComp,
    CaseConnectInToWireComp,
    CaseConnectSliceToOutComp,
    CasePartSelOverBitSelComp,
    CasePartSelOverPartSelComp,
)
from ..VStructuralTranslatorL1 import VStructuralTranslatorL1


def run_test( case, m ):
  m.elaborate()
  VStructuralTranslatorL1.is_verilog_reserved = lambda s, x: x in verilog_reserved
  tr = VStructuralTranslatorL1( m )
  tr.clear( m )
  tr._rtlir_tr_unpacked_q = deque()
  tr.translate_structural( m )

  ports = tr.structural.decl_ports[m]
  wires = tr.structural.decl_wires[m]
  consts = tr.structural.decl_consts[m]
  conns = tr.structural.connections[m]

  check_eq( ports,  case.REF_PORT  )
  check_eq( wires,  case.REF_WIRE  )
  check_eq( consts, case.REF_CONST )
  check_eq( conns,  case.REF_CONN  )

  try:
    placeholder_wrapper = tr.structural.placeholder_wrapper[m]
    check_eq( placeholder_wrapper, case.REF_PLACEHOLDER_WRAPPER )

    placeholder_dependency = tr.structural.placeholder_dependency[m]
    check_eq( placeholder_dependency, case.REF_PLACEHOLDER_DEPENDENCY )
  except AttributeError:
    pass

@pytest.mark.parametrize(
  'case', [
    CaseConnectInToWireComp,
    CaseConnectBitsConstToOutComp,
    CaseConnectConstToOutComp,
    CaseConnectBitSelToOutComp,
    CaseConnectSliceToOutComp,
    CaseBitSelOverBitSelComp,
    CaseBitSelOverPartSelComp,
    CasePartSelOverBitSelComp,
    CasePartSelOverPartSelComp,
  ]
)
def test_verilog_structural_L1( case ):
  run_test( case, case.DUT() )



"""
==========================================================================
bitstruct.py
==========================================================================
APIs to generate a bitstruct type. Using decorators and type annotations
to create bit struct is much inspired by python3 dataclass implementation.
Note that the implementation (such as the _CAPITAL constants to add some
private metadata) in this file is very similar to the **original python3
dataclass implementation**. The syntax of creating bit struct is very
similar to that of python3 dataclass.
https://github.com/python/cpython/blob/master/Lib/dataclasses.py

For example,

@bitstruct
class Point:
  x : Bits4
  y : Bits4

will automatically generate some methods, such as __init__, __str__,
__repr__, for the Point class.

Similar to the built-in dataclasses module, we also provide a
mk_bitstruct function for user to dynamically generate bit struct types.
For example,

mk_bitstruct( 'Pixel',{
    'r' : Bits4,
    'g' : Bits4,
    'b' : Bits4,
  },
  name_space = {
    '__str__' : lambda self: f'({self.r},{self.g},{self.b})'
  }
)

is equivalent to:

@bitstruct
class Pixel:
  r : Bits4
  g : Bits4
  b : Bits4

  def __str__( self ):
    return f'({self.r},{self.g},{self.b})'

Author : Yanghui Ou, Shunning Jiang
  Date : Oct 19, 2019
"""
import functools
import keyword
import operator
import types
import warnings

import py

from pymtl3.extra.pypy import custom_exec

from .bits_import import *
from .helpers import concat

#-------------------------------------------------------------------------
# Constants
#-------------------------------------------------------------------------

# Object with this attribute is considered as bit struct, as we assume
# only the bitstruct decorator will stamp this attribute to a class. This
# attribute also stores the field information and can be used for
# translation.
#
# The original dataclass use hasattr( cls, _FIELDS ) to check dataclass.
# We do this here as well

_FIELDS = '__bitstruct_fields__'

def is_bitstruct_inst( obj ):
  """Returns True if obj is an instance of a dataclass."""
  return hasattr(type(obj), _FIELDS)

def is_bitstruct_class(cls):
  """Returns True if obj is a dataclass ."""
  return isinstance(cls, type) and hasattr(cls, _FIELDS)

def get_bitstruct_inst_all_classes( obj ):
  # list: put all types together
  if isinstance( obj, list ):
    return functools.reduce( operator.or_, [ get_bitstruct_inst_all_classes(x) for x in obj ] )
  ret = { obj.__class__ }
  # BitsN or int
  if isinstance( obj, (Bits, int) ):
    return ret
  # BitStruct
  assert is_bitstruct_inst( obj ), f"{obj} is not a valid PyMTL Bitstruct!"
  return ret | functools.reduce( operator.or_, [ get_bitstruct_inst_all_classes(getattr(obj, v))
                                                for v in obj.__bitstruct_fields__.keys() ] )

_DEFAULT_SELF_NAME = 's'
_ANTI_CONFLICT_SELF_NAME = '__bitstruct_self__'

#-------------------------------------------------------------------------
# _create_fn
#-------------------------------------------------------------------------
# A helper function that creates a function based on
# - fn_name : name of the function
# - args_lst : a list of arguments in string
# - body_lst : a list of statement of the function body in string

# Also note that this whole _create_fn thing is similar to the original
# dataclass implementation!

def _create_fn( fn_name, args_lst, body_lst, _globals=None ):
  # Assemble argument string and body string
  args = ', '.join(args_lst)
  body = '\n'.join(f'  {statement}' for statement in body_lst)

  # Assemble the source code and execute it
  src = f'def {fn_name}({args}):\n{body}'
  if _globals is None: _globals = {}
  _locals = {}
  custom_exec( py.code.Source(src).compile(), _globals, _locals )
  return _locals[fn_name]

#-------------------------------------------------------------------------
# _mk_init_arg
#-------------------------------------------------------------------------
# Creates a init argument string from a field.
#
# Shunning: I revamped the whole thing because they are indeed mutable
# objects.

def _mk_init_arg( name, type_ ):
  # default is always None
  if isinstance( type_, list ) or is_bitstruct_class( type_ ):
    return f'{name} = None'
  return f'{name} = 0'

#-------------------------------------------------------------------------
# _mk_init_body
#-------------------------------------------------------------------------
# Creates one line of __init__ body from a field
# to globals.

def _mk_init_body( self_name, name, type_ ):
  def _recursive_generate_init( x ):
    if isinstance( x, list ):
      return f"[{', '.join( [ _recursive_generate_init(x[0]) ] * len(x) )}]"
    return f"_type_{name}()"

  if isinstance( type_, list ) or is_bitstruct_class( type_ ):
    return f'{self_name}.{name} = {name} or {_recursive_generate_init(type_)}'

  assert issubclass( type_, Bits )
  return f'{self_name}.{name} = _type_{name}({name})'

#-------------------------------------------------------------------------
# _mk_tuple_str
#-------------------------------------------------------------------------
# Creates a tuple of string representations of each field. For example,
# if the self_name is 'self' and fields is [ 'x', 'y' ], it will return
# ('self.x', 'self.y'). This is used for creating the default __eq__ and
# __hash__ function.

def _mk_tuple_str( self_name, fields ):
  return f'({",".join([f"{self_name}.{name}" for name in fields])},)'

#-------------------------------------------------------------------------
# _mk_init_fn
#-------------------------------------------------------------------------
# Creates a __init__ function based on fields. For example, if fields
# contains two field x (Bits4) and y (Bits4), _mk_init_fn will return a
# function that looks like the following:
#
# def __init__( s, x = 0, y = 0, z = None, p = None ):
#   s.x = _type_x(x)
#   s.y = _type_y(y)
#   s.z = z or _type_z()
#   s.p = p or [ _type_p(), _type_p() ]
#
# NOTE:
# _mk_init_fn also takes as argument the name of self in case there is a
# field with name 's' or 'self'.
#
# TODO: should we provide a __post_init__ function like dataclass does?

def _mk_init_fn( self_name, fields ):

  # Register necessary types in _globals
  _globals = {}

  for name, type_ in fields.items():
    if isinstance( type_, list ):
      x = type_[0]
      while isinstance( x, list ):
        x = x[0]
      _globals[ f"_type_{name}" ] = x
    else:
      assert issubclass( type_, Bits ) or is_bitstruct_class( type_ )
      _globals[ f"_type_{name}" ] = type_

  return _create_fn(
    '__init__',
    [ self_name ] + [ _mk_init_arg( *field ) for field in fields.items() ],
    [ _mk_init_body( self_name, *field ) for field in fields.items() ],
    _globals = _globals,
  )

#-------------------------------------------------------------------------
# _mk_str_fn
#-------------------------------------------------------------------------
# Creates a __str__ function based on fields. For example, if fields
# contains two field x (Bits4) and y (Bits4), _mk_str_fn will return a
# function that looks like the following:
#
# def __str__( self ):
#   return f'{self.x}:{self.y}'

def _mk_str_fn( fields ):
  return _create_fn(
    '__str__',
    [ 'self' ],
    [ 'return f"' +
      ':'.join([ f'{{self.{name}}}' for name in fields ]) + '"']
  )

#-------------------------------------------------------------------------
# _mk_repr_fn
#-------------------------------------------------------------------------
# Creates a __repr__ function based on fields. For example, if fields
# contains two field x (Bits4) and y (Bits4), _mk_repr_fn will return a
# function that looks like the following:
#
# def __repr__( self ):
#  return self.__class__.__name__ + f'(x={self.x!r}, y={self.y!r})'

def _mk_repr_fn( fields ):
  return _create_fn(
    '__repr__',
    [ 'self' ],
    [ 'return self.__class__.__name__ + f"(' +
      ','.join([ f'{{self.{name}!r}}' for name in fields ]) +
      ')"']
  )

#-------------------------------------------------------------------------
# _mk_eq_fn
#-------------------------------------------------------------------------
# Creates a __eq__ function based on fields. By default it just compares
# each field. For example, if fields contains two field x (Bits4) and y
# (Bits4), _mk_eq_fn will return a function that looks like the
# following:
#
# def __eq__( self, other ):
#   if other.__class__ is self.__class__:
#     return (self.x,self.y,) == (other.x,other.y,)
#   else:
#     raise NotImplemented

def _mk_eq_fn( fields ):
  self_tuple  = _mk_tuple_str( 'self', fields )
  other_tuple = _mk_tuple_str( 'other', fields )
  return _create_fn(
    '__eq__',
    [ 'self', 'other' ],
    [ f'return (other.__class__ is self.__class__) and {self_tuple} == {other_tuple}' ]
  )

#-------------------------------------------------------------------------
# _mk_hash_fn
#-------------------------------------------------------------------------
# Creates a __hash__ function based on fields. By default it just hashes
# all fields. For example, if fields contains two field x (Bits4) and y
# (Bits4), _mk_hash_fn will return a function that looks like the
# following:
#
# def __hash__( self ):
#   return hash((self.x,self.y,))

def _mk_hash_fn( fields ):
  self_tuple = _mk_tuple_str( 'self', fields )
  return _create_fn(
    '__hash__',
    [ 'self' ],
    [ f'return hash({self_tuple})' ]
  )

#--------------------------PyMTL3 specific--------------------------------

#-------------------------------------------------------------------------
# _mk_ff_fn
#-------------------------------------------------------------------------
# Creates __ilshift__ and _flip functions that looks like the following:
#
# def __ilshift__( self, other ):
#   if self.__class__ is not other.__class__:
#     other = self.__class__.from_bits( other.to_bits() )
#   self.x <<= other.x
#   self.y[0][0] <<= other.y[0][0]
#
# def _flip( self ):
#   self.x._flip()
#   self.y[i][j]._flip()

def _mk_ff_fn( fields ):

  def _gen_list_ilshift_strs( type_, prefix='' ):
    if isinstance( type_, list ):
      ilshift_strs, flip_strs = [], []
      for i in range(len(type_)):
        ils, fls = _gen_list_ilshift_strs( type_[0], f"{prefix}[{i}]" )
        ilshift_strs.extend( ils )
        flip_strs.extend( fls )
      return ilshift_strs, flip_strs
    else:
      return [ f"self.{prefix} <<= other.{prefix}" ], [f"self.{prefix}._flip()"]

  ilshift_strs = [ 'if self.__class__ is not other.__class__:',
                   '  other = self.__class__.from_bits( other.to_bits() )']
  flip_strs = []
  for name, type_ in fields.items():
    ils, fls = _gen_list_ilshift_strs( type_, name )
    ilshift_strs.extend( ils )
    flip_strs.extend( fls )

  return _create_fn(
    '__ilshift__',
    [ 'self', 'other' ],
    ilshift_strs + [ "return self" ],
  ), _create_fn(
    '_flip',
    [ 'self' ],
    flip_strs,
  ),

#-------------------------------------------------------------------------
# _mk_clone_fn
#-------------------------------------------------------------------------
# Creates clone function that looks like the following:
# Use this clone function in any place that you need to perform a
# deepcopy on a bitstruct.
#
# def clone( self ):
#   return self.__class__( self.x.clone(), [ self.y[0].clone(), self.y[1].clone() ]  )

def _gen_list_clone_strs( type_, prefix='' ):
  if isinstance( type_, list ):
    return "[" + ",".join( [ _gen_list_clone_strs( type_[0], f"{prefix}[{i}]" )
                        for i in range(len(type_)) ] ) + "]"
  else:
    return f"{prefix}.clone()"

def _mk_clone_fn( fields ):
  clone_strs = [ 'return self.__class__(' ]

  for name, type_ in fields.items():
    clone_strs.append( "  " + _gen_list_clone_strs( type_, f'self.{name}' ) + "," )

  return _create_fn(
    'clone',
    [ 'self' ],
    clone_strs + [ ')' ],
  )

def _mk_deepcopy_fn( fields ):
  clone_strs = [ 'return self.__class__(' ]

  for name, type_ in fields.items():
    clone_strs.append( "  " + _gen_list_clone_strs( type_, f'self.{name}' ) + "," )

  return _create_fn(
    '__deepcopy__',
    [ 'self', 'memo' ],
    clone_strs + [ ')' ],
  )

#-------------------------------------------------------------------------
# _mk_imatmul_fn
#-------------------------------------------------------------------------
# Creates @= function that copies the value over ...
# TODO create individual from_bits for imatmul and ilshift

# def __imatmul__( self, other ):
#   if self.__class__ is not other.__class__:
#     other = self.__class__.from_bits( other.to_bits() )
#   self.x @= other.x
#   self.y[0] @= other.y[0]
#   self.y[1] @= other.y[1]

def _mk_imatmul_fn( fields ):

  def _gen_list_imatmul_strs( type_, prefix='' ):
    if isinstance( type_, list ):
      ret = []
      for i in range(len(type_)):
        ret.extend( _gen_list_imatmul_strs( type_[0], f"{prefix}[{i}]" ) )
      return ret
    else:
      return [ f"self.{prefix} @= other.{prefix}" ]

  imatmul_strs = [ 'if self.__class__ is not other.__class__:',
                   '  other = self.__class__.from_bits( other.to_bits() )']
  for name, type_ in fields.items():
    imatmul_strs.extend( _gen_list_imatmul_strs( type_, name ) )

  return _create_fn(
    '__imatmul__',
    [ 'self', 'other' ],
    imatmul_strs + [ "return self" ],
  )

#-------------------------------------------------------------------------
# _mk_nbits_to_bits_fn
#-------------------------------------------------------------------------
# Creates nbits, to_bits function that copies the value over ...
#
# def to_bits( self ):
#   return concat( self.x, self.y[0], self.y[1] )
#
# TODO packing order of array? x[0] is LSB or MSB of a list
# current we do LSB

def _mk_nbits_to_bits_fn( fields ):

  def _gen_to_bits_strs( type_, prefix, start_bit ):

    if isinstance( type_, list ):
      to_strs = []
      # The packing order is LSB, so we need to reverse the list to make x[-1] higher bits
      for i in reversed(range(len(type_))):
        start_bit, tos = _gen_to_bits_strs( type_[0], f"{prefix}[{i}]", start_bit )
        to_strs.extend( tos )
      return start_bit, to_strs

    elif is_bitstruct_class( type_ ):
      to_strs = []
      for name, typ in getattr(type_, _FIELDS).items():
        start_bit, tos = _gen_to_bits_strs( typ, f"{prefix}.{name}", start_bit )
        to_strs.extend( tos )
      return start_bit, to_strs

    else:
      end_bit = start_bit + type_.nbits
      return end_bit, [ f"self.{prefix}" ]

  to_bits_strs = []
  total_nbits  = 0
  for name, type_ in fields.items():
    total_nbits, tos = _gen_to_bits_strs( type_, name, total_nbits )
    to_bits_strs.extend( tos )

  return total_nbits, _create_fn( 'to_bits', [ 'self' ],
                                  [ f"return concat({', '.join(to_bits_strs)})" ],
                                  _globals={'concat':concat} )

#-------------------------------------------------------------------------
# _mk_from_bits_fn
#-------------------------------------------------------------------------
# Creates static method from_bits that creates a new bitstruct based on Bits
# and instance method _from_bits that copies the value over
#
# @staticmethod
# def from_bits( other ):
#   return self.__class__( other[16:32], other[0:16] )

def _mk_from_bits_fns( fields, total_nbits ):

  def _gen_from_bits_strs( type_, end_bit ):

    if isinstance( type_, list ):
      from_strs = []
      # Since we are doing LSB for x[0], we need to unpack from the last
      # element of the list, and then reverse it again to construct a list ...
      for i in range(len(type_)):
        end_bit, fs = _gen_from_bits_strs( type_[0], end_bit )
        from_strs.extend( fs )
      return end_bit, [ f"[{','.join(reversed(from_strs))}]" ]

    elif is_bitstruct_class( type_ ):
      if type_ in type_name_mapping:
        type_name = type_name_mapping[ type_ ]
      else:
        type_name = f"_type{len(type_name_mapping)}"
        type_name_mapping[ type_ ] = type_name

      from_strs = []
      for name, typ in getattr(type_, _FIELDS).items():
        end_bit, fs = _gen_from_bits_strs( typ, end_bit )
        from_strs.extend( fs )
      return end_bit, [ f"{type_name}({','.join(from_strs)})" ]

    else:
      if type_ not in type_name_mapping:
        type_name_mapping[ type_ ] = type_.__name__
      else:
        assert type_name_mapping[ type_ ] == type_.__name__
      start_bit = end_bit - type_.nbits
      return start_bit, [ f"other[{start_bit}:{end_bit}]" ]

  from_bits_strs = []
  end_bit = total_nbits

  # This is to make sure we capture two types with the same name but different
  # attributes
  type_name_mapping = {}
  type_count = 0

  for _, type_ in fields.items():
    end_bit, fs = _gen_from_bits_strs( type_, end_bit )
    from_bits_strs.extend( fs )

  assert end_bit == 0
  _globals = { y: x for x,y in type_name_mapping.items() }
  assert len(_globals) == len(type_name_mapping)

  # TODO add assertion in bits
  return _create_fn( 'from_bits', [ 'cls', 'other' ],
                     [ "assert cls.nbits == other.nbits, f'LHS bitstruct {cls.nbits}-bit <> RHS other {other.nbits}-bit'",
                       "other = other.to_bits()",
                       f"return cls({','.join(from_bits_strs)})" ], _globals )
#-------------------------------------------------------------------------
# _check_valid_array
#-------------------------------------------------------------------------

def _recursive_check_array_types( current ):
  x = current[0]
  if isinstance( x, list ):
    x_len  = len(x)
    x_type = _recursive_check_array_types( x )
    for y in current[1:]:
      assert isinstance( y, list ) and len(y) == x_len
      y_type = _recursive_check_array_types( y )
      assert y_type is x_type
    return x_type

  assert issubclass( x, Bits ) or is_bitstruct_class( x )
  for y in current[1:]:
    assert y is x
  return x

def _check_valid_array_of_types( arr ):
  # Check if the provided list is a strict multidimensional array
  try:
    return _recursive_check_array_types( arr )
  except Exception as e:
    print(e)
    return None

#-------------------------------------------------------------------------
# _check_field_annotation
#-------------------------------------------------------------------------

def _check_field_annotation( cls, name, type_ ):
  # Make sure not default is annotated
  if hasattr( cls, name ):
    default = getattr( cls, name )
    raise TypeError( "We don't allow subfields to have default value:\n"
                     f"- Field '{name}' of BitStruct {cls.__name__} has default value {default!r}." )

  # Special case if the type is an instance of list
  if isinstance( type_, list ):
    if _check_valid_array_of_types( type_ ) is None:
      raise TypeError( "The provided list spec should be a strict multidimensional ARRAY "
                       "with no varying sizes or types. All non-list elements should be VALID types." )
  else:
    # Now we work with types
    if not isinstance( type_, type ):
      raise TypeError(f"{type_} is not a type\n"\
                      f"- Field '{name}' of BitStruct {cls.__name__} is annotated as {type_}.")

    # More specifically, Bits and BitStruct
    if not issubclass( type_, Bits ) and not is_bitstruct_class( type_ ):
      raise TypeError( "We currently only support BitsN, list, or another BitStruct as BitStruct field:\n"
                      f"- Field '{name}' of BitStruct {cls.__name__} is annotated as {type_}." )

#-------------------------------------------------------------------------
# _get_self_name
#-------------------------------------------------------------------------
# Return a self name based on fields.

def _get_self_name( fields ):
  return( _ANTI_CONFLICT_SELF_NAME if _DEFAULT_SELF_NAME in fields else
          _DEFAULT_SELF_NAME )

#-------------------------------------------------------------------------
# _process_cls
#-------------------------------------------------------------------------
# Process the input cls and add methods to it.

_bitstruct_hash_cache = {}

def _process_class( cls, add_init=True, add_str=True, add_repr=True,
                    add_hash=True ):

  # Get annotations of the class
  cls_annotations = cls.__dict__.get('__annotations__', {})
  if not cls_annotations:
    raise AttributeError( "No field is declared in the bit struct definition.\n"
                         f"Suggestion: check the definition of {cls.__name__} to"
                          " make sure it only contains 'field_name(string): Type(type).'" )

  # Get field information from the annotation and prepare for hashing
  fields = {}
  hashable_fields = {}

  def _convert_list_to_tuple( x ):
    if isinstance( x, list ):
      return tuple( [ _convert_list_to_tuple( y ) for y in x ] )
    return x

  reserved_fields = ['to_bits', 'from_bits', 'nbits']
  for x in reserved_fields:
    assert x not in cls.__dict__, f"Currently a bitstruct cannot have {reserved_fields}, but "\
                                  f"{x} is provided as {cls.__dict__[x]}"


  for a_name, a_type in cls_annotations.items():
    assert a_name not in reserved_fields, f"Currently a bitstruct cannot have {reserved_fields}, but "\
                                          f"{a_name} is annotated as {a_type}"
    _check_field_annotation( cls, a_name, a_type )
    fields[ a_name ] = a_type
    hashable_fields[ a_name ] = _convert_list_to_tuple( a_type )

  cls._hash = _hash = hash( (cls.__name__, *tuple(hashable_fields.items()),
                             add_init, add_str, add_repr, add_hash) )

  if _hash in _bitstruct_hash_cache:
    return _bitstruct_hash_cache[ _hash ]

  _bitstruct_hash_cache[ _hash ] = cls

  # Stamp the special attribute so that translation pass can identify it
  # as bit struct.
  setattr( cls, _FIELDS, fields )

  # Add methods to the class

  # Create __init__. Here I follow the dataclass convention that we only
  # add our generated __init__ function when add_init is true and user
  # did not define their own init.
  if add_init:
    if not '__init__' in cls.__dict__:
      cls.__init__ = _mk_init_fn( _get_self_name(fields), fields )

  # Create __str__
  if add_str:
    if not '__str__' in cls.__dict__:
      cls.__str__ = _mk_str_fn( fields )

  # Create __repr__
  if add_repr:
    if not '__repr__' in cls.__dict__:
      cls.__repr__ = _mk_repr_fn( fields )

  # Create __eq__. There is no need for a __ne__ method as python will
  # call __eq__ and negate it.
  # NOTE: if user overwrites __eq__ it may lead to different behavior for
  # the translated verilog as in the verilog world two bit structs are
  # equal only if all the fields are equal. We always try to add __eq__

  if not '__eq__' in cls.__dict__:
    cls.__eq__ = _mk_eq_fn( fields )
  else:
    w_msg = ( f'Overwriting {cls.__qualname__}\'s __eq__ may cause the '
              'translated verilog behaves differently from PyMTL '
              'simulation.')
    warnings.warn( w_msg )

  # Create __hash__.
  if add_hash:
    if not '__hash__' in cls.__dict__:
      cls.__hash__ = _mk_hash_fn( fields )

  # Shunning: add __ilshift__ and _flip for update_ff
  assert not '__ilshift__' in cls.__dict__ and not '_flip' in cls.__dict__

  cls.__ilshift__, cls._flip = _mk_ff_fn( fields )

  # Shunning: add clone
  assert not 'clone' in cls.__dict__ and not '__deepcopy__' in cls.__dict__

  cls.clone = _mk_clone_fn( fields )

  cls.__deepcopy__ = _mk_deepcopy_fn( fields )

  # Shunning: add imatmul for assignment, as well as nbits/to_bits/from_bits
  assert '__imatmul__' not in cls.__dict__ and 'to_bits' not in cls.__dict__ and \
         'nbits' not in cls.__dict__ and 'from_bits' not in cls.__dict__

  cls.__imatmul__ = _mk_imatmul_fn( fields )
  cls.nbits, cls.to_bits = _mk_nbits_to_bits_fn( fields )

  from_bits = _mk_from_bits_fns( fields, cls.nbits )
  cls.from_bits = classmethod(from_bits)

  assert not 'get_field_type' in cls.__dict__

  def get_field_type( cls, name ):
    if name in cls.__bitstruct_fields__:
      return cls.__bitstruct_fields__[ name ]
    raise AttributeError( f"{cls} has no field '{name}'" )

  cls.get_field_type = classmethod(get_field_type)

  # TODO: maybe add a to_bits and from bits function.

  return cls

#-------------------------------------------------------------------------
# bitstruct
#-------------------------------------------------------------------------
# The actual class decorator. We add a * in the argument list so that the
# following argument can only be used as keyword arguments.

def bitstruct( _cls=None, *, add_init=True, add_str=True, add_repr=True, add_hash=True ):

  def wrap( cls ):
    return _process_class( cls, add_init, add_str, add_repr )

  # Called as @bitstruct(...)
  if _cls is None:
    return wrap

  # Called as @bitstruct without parens.
  return wrap( _cls )

#-------------------------------------------------------------------------
# mk_bitstruct
#-------------------------------------------------------------------------
# Dynamically generate a bit struct class.
# TODO: should we add base parameters to support inheritence?

def mk_bitstruct( cls_name, fields, *, namespace=None, add_init=True,
                   add_str=True, add_repr=True, add_hash=True ):

  # copy namespace since  will mutate it
  namespace = {} if namespace is None else namespace.copy()

  # We assume fields is a dictionary and thus there won't be duplicate
  # field names. So we only check if the field names are indeed strings
  # and that they are not keywords.
  annos = {}
  for name, f in fields.items():
    if not isinstance( name, str ) or not name.isidentifier():
      raise TypeError( f'Field name {name!r} is not a valid identifier!' )
    if keyword.iskeyword( name ):
      raise TypeError( f'Field name {name!r} is a keyword!' )
    annos[ name ] = f

  namespace['__annotations__'] = annos
  cls = types.new_class( cls_name, (), {}, lambda ns: ns.update( namespace ) )
  return bitstruct( cls, add_init=add_init, add_str=add_str,
                    add_repr=add_repr, add_hash=add_hash )



#=========================================================================
# elf
#=========================================================================
# A simple translator between ELF files and a sparse memory image object.
# Note that the translator is far from complete but is sufficient for use
# in our research processors. I found this document pretty useful for
# understanding the ELF32 format:
#
#  http://docs.oracle.com/cd/E19457-01/801-6737/801-6737.pdf
#
# Note that this implementation is inspired by the ELF object file reader
# here:
#
#  http://www.tinyos.net/tinyos-1.x/tools/src/mspgcc-pybsl/elf.py
#
# which includes this copyright:
#
#  (C) 2003 cliechti@gmx.net
#  Python license
#
# Shunning: I ported it to Python3 today and it is a bit interesting due
# to the recent bytes/str disambiguation in Python3. Basically, there
# are several changes worth noting:
# - Currently file_obj.read() will return _bytes_ instead of str, so if
#   we want to use the variable as string, we need to perform
#   bytes.decode().
# - struct.unpack only works for string/bytearray (I guess?) so I had to
#   use bytearray(bytes_obj) to avoid any changes to struct.unpack and the
#   format string.
# - Whenever we write the elf file, we have to convert the strings back
#   to bytes. Just keep in mind that for string purposes we need to decode
#   and encode back.
#
# Author : Christopher Batten, Shunning Jiang
# Date   : Feb 26, 2020

import struct

from .SparseMemoryImage import SparseMemoryImage

#-------------------------------------------------------------------------
# ELF File Format Types
#-------------------------------------------------------------------------
# These are the sizes for various ELF32 data types used in describing
# various structures below.
#
#            size alignment
# elf_addr   4    4  Unsigned program address
# elf_half   2    2  Unsigned medium integer
# elf_off    4    4  Unsigned file offset
# elf_sword  4    4  Signed   large integer
# elf_word   4    4  Unsigned large integer
# elf_byte   1    1  Unsigned small integer

#=========================================================================
# ElfHeader
#=========================================================================
# Class encapsulating an ELF32 header which implements the following
# C-structure.
#
# define EI_NIDENT 16
# typedef struct {
#   elf_byte e_ident[EI_NIDENT];
#   elf_half e_type;
#   elf_half e_machine;
#   elf_word e_version;
#   elf_addr e_entry;
#   elf_off  e_phoff;
#   elf_off  e_shoff;
#   elf_word e_flags;
#   elf_half e_ehsize;
#   elf_half e_phentsize;
#   elf_half e_phnum;
#   elf_half e_shentsize;
#   elf_half e_shnum;
#   elf_half e_shstrndx;
# } elf_ehdr;

class ElfHeader:

  FORMAT = "<16sHHIIIIIHHHHHH"
  NBYTES = struct.calcsize( FORMAT )

  # Offsets within e_ident

  IDENT_NBYTES      = 16     # Size of e_ident[]
  IDENT_IDX_MAG0    = 0      # File identification
  IDENT_IDX_MAG1    = 1      # File identification
  IDENT_IDX_MAG2    = 2      # File identification
  IDENT_IDX_MAG3    = 3      # File identification
  IDENT_IDX_CLASS   = 4      # File class
  IDENT_IDX_DATA    = 5      # Data encoding
  IDENT_IDX_VERSION = 6      # File version
  IDENT_IDX_PAD     = 7      # Start of padding bytes

  # Elf file type flags

  TYPE_NONE         = 0      # No file type
  TYPE_REL          = 1      # Relocatable file
  TYPE_EXEC         = 2      # Executable file
  TYPE_DYN          = 3      # Shared object file
  TYPE_CORE         = 4      # Core file
  TYPE_LOPROC       = 0xff00 # Processor-specific
  TYPE_HIPROC       = 0xffff # Processor-specific

  #-----------------------------------------------------------------------
  # Constructor
  #-----------------------------------------------------------------------

  def __init__( self, data=None ):
    if data != None:
      self.from_bytes( data )

  #-----------------------------------------------------------------------
  # from_bytes
  #-----------------------------------------------------------------------

  def from_bytes( self, data ):
    ehdr_list = struct.unpack( ElfHeader.FORMAT, bytearray(data) )
    self.ident     = ehdr_list[0].decode()
    self.type      = ehdr_list[1]
    self.machine   = ehdr_list[2]
    self.version   = ehdr_list[3]
    self.entry     = ehdr_list[4]
    self.phoff     = ehdr_list[5]
    self.shoff     = ehdr_list[6]
    self.flags     = ehdr_list[7]
    self.ehsize    = ehdr_list[8]
    self.phentsize = ehdr_list[9]
    self.phnum     = ehdr_list[10]
    self.shentsize = ehdr_list[11]
    self.shnum     = ehdr_list[12]
    self.shstrndx  = ehdr_list[13]

  #-----------------------------------------------------------------------
  # to_bytes
  #-----------------------------------------------------------------------

  def to_bytes( self ):
    return struct.pack( ElfHeader.FORMAT,
      self.ident.encode(),
      self.type,
      self.machine,
      self.version,
      self.entry,
      self.phoff,
      self.shoff,
      self.flags,
      self.ehsize,
      self.phentsize,
      self.phnum,
      self.shentsize,
      self.shnum,
      self.shstrndx,
   )

  #-----------------------------------------------------------------------
  # __str__
  #-----------------------------------------------------------------------

  def __str__( self ):
    return \
"""
 ElfHeader:
   ident     = {},
   type      = {},
   machine   = {},
   version   = {},
   entry     = {},
   phoff     = {},
   shoff     = {},
   flags     = {},
   ehsize    = {},
   phentsize = {},
   phnum     = {},
   shentsize = {},
   shnum     = {},
   shstrndx  = {}
""".format(
   self.ident,
   self.type,
   self.machine,
   self.version,
   hex(self.entry),
   hex(self.phoff),
   hex(self.shoff),
   hex(self.flags),
   self.ehsize,
   self.phentsize,
   self.phnum,
   self.shentsize,
   self.shnum,
   self.shstrndx,
)

#=========================================================================
# ElfSectionHeader
#=========================================================================
# Class encapsulating an ELF32 section header which implements the
# following C-structure.
#
# typedef struct {
#   elf_word sh_name;
#   elf_word sh_type;
#   elf_word sh_flags;
#   elf_addr sh_addr;
#   elf_off  sh_offset;
#   elf_word sh_size;
#   elf_word sh_link;
#   elf_word sh_info;
#   elf_word sh_addralign;
#   elf_word sh_entsize;
# } elf_shdr;
#

class ElfSectionHeader:

  FORMAT = "<IIIIIIIIII"
  NBYTES = struct.calcsize( FORMAT )

  # Section types. Note that we only load some of these sections.

  TYPE_NULL        = 0
  TYPE_PROGBITS    = 1 # \
  TYPE_SYMTAB      = 2 # | We only load sections of these types
  TYPE_STRTAB      = 3 # /
  TYPE_RELA        = 4
  TYPE_HASH        = 5
  TYPE_DYNAMIC     = 6
  TYPE_NOTE        = 7
  TYPE_NOBITS      = 8
  TYPE_REL         = 9
  TYPE_SHLIB       = 10
  TYPE_DYNSYM      = 11
  TYPE_LOPROC      = 0x70000000
  TYPE_HIPROC      = 0x7fffffff
  TYPE_LOUSER      = 0x80000000
  TYPE_HIUSER      = 0xffffffff

  # Section attribute flags. Note that we only load sections with the
  # SHF_ALLOC flag set into the actual sparse memory.

  FLAGS_WRITE       = 0x1
  FLAGS_ALLOC       = 0x2
  FLAGS_EXECINSTR   = 0x4
  FLAGS_MASKPROC    = 0xf0000000

  #-----------------------------------------------------------------------
  # Constructor
  #-----------------------------------------------------------------------

  def __init__( self, data=None ):
    if data != None:
      self.from_bytes( data )

  #-----------------------------------------------------------------------
  # from_bytes
  #-----------------------------------------------------------------------

  def from_bytes( self, data ):
    shdr_list = struct.unpack( ElfSectionHeader.FORMAT, bytearray(data) )
    self.name      = shdr_list[0]
    self.type      = shdr_list[1]
    self.flags     = shdr_list[2]
    self.addr      = shdr_list[3]
    self.offset    = shdr_list[4]
    self.size      = shdr_list[5]
    self.link      = shdr_list[6]
    self.info      = shdr_list[7]
    self.addralign = shdr_list[8]
    self.entsize   = shdr_list[9]

  #-----------------------------------------------------------------------
  # to_bytes
  #-----------------------------------------------------------------------

  def to_bytes( self ):
    return struct.pack( ElfSectionHeader.FORMAT,
      self.name,
      self.type,
      self.flags,
      self.addr,
      self.offset,
      self.size,
      self.link,
      self.info,
      self.addralign,
      self.entsize,
   )

  #-----------------------------------------------------------------------
  # __str__
  #-----------------------------------------------------------------------

  def __str__( self ):
    return \
"""
 ElfSectionHeader:
   name      = {},
   type      = {},
   flags     = {},
   addr      = {},
   offset    = {},
   size      = {},
   link      = {},
   info      = {},
   addralign = {},
   entsize   = {},
""".format(
   self.name,
   self.type,
   hex(self.flags),
   hex(self.addr),
   hex(self.offset),
   self.size,
   self.link,
   self.info,
   self.addralign,
   self.entsize,
)

#=========================================================================
# ElfSymTabEntry
#=========================================================================
# Class encapsulating an ELF32 symbol table entry which implements the
# following C-structure.
#
# typedef struct {
#   elf_word st_name;
#   elf_addr st_value;
#   elf_word st_size;
#   elf_byte st_info;
#   elf_byte st_other;
#   elf_half st_shndx;
# } elf_sym;
#

class ElfSymTabEntry:

  FORMAT = "<IIIBBH"
  NBYTES = struct.calcsize( FORMAT )

  # Symbol types. Note we only load some of these types.

  TYPE_NOTYPE  = 0 # \
  TYPE_OBJECT  = 1 # | We only load symbols of these types
  TYPE_FUNC    = 2 # /
  TYPE_SECTION = 3
  TYPE_FILE    = 4
  TYPE_LOPROC  = 13
  TYPE_HIPROC  = 15

  #-----------------------------------------------------------------------
  # Constructor
  #-----------------------------------------------------------------------

  def __init__( self, data=None ):
    if data != None:
      self.from_bytes( data )

  #-----------------------------------------------------------------------
  # from_bytes
  #-----------------------------------------------------------------------

  def from_bytes( self, data ):
    sym_list = struct.unpack( ElfSymTabEntry.FORMAT, bytearray(data) )
    self.name  = sym_list[0]
    self.value = sym_list[1]
    self.size  = sym_list[2]
    self.info  = sym_list[3]
    self.other = sym_list[4]
    self.shndx = sym_list[5]

  #-----------------------------------------------------------------------
  # to_bytes
  #-----------------------------------------------------------------------

  def to_bytes( self ):
    return struct.pack( ElfSymTabEntry.FORMAT,
      self.name,
      self.value,
      self.size,
      self.info,
      self.other,
      self.shndx,
   )

  #-----------------------------------------------------------------------
  # __str__
  #-----------------------------------------------------------------------

  def __str__( self ):
    return \
"""
 ElfSymTabEntry:
   ident     = {}
   value     = {}
   size      = {}
   info      = {}
   other     = {}
   shndx     = {}
""".format(
   self.name,
   hex(self.value),
   self.size,
   self.info,
   self.other,
   self.shndx,
)

#-------------------------------------------------------------------------
# elf_reader
#-------------------------------------------------------------------------
# Opens and parses an ELF file into a sparse memory image object.

def elf_reader( file_obj ):

  # Read the data for the ELF header

  ehdr_data = file_obj.read( ElfHeader.NBYTES )

  # Construct an ELF header object

  ehdr = ElfHeader( ehdr_data )

  # Verify if its a known format and really an ELF file

  if ehdr.ident[0:4] != '\x7fELF':
    raise ValueError( "Not a valid ELF file" )

  # We need to find the section string table so we can figure out the
  # name of each section. We know that the section header for the section
  # string table is entry shstrndx, so we first get the data for this
  # section header.

  file_obj.seek( ehdr.shoff + ehdr.shstrndx * ehdr.shentsize )
  shdr_data = file_obj.read(ehdr.shentsize)

  # Construct a section header object for the section string table

  shdr = ElfSectionHeader( shdr_data )

  # Read the data for the section header table

  file_obj.seek( shdr.offset )
  shstrtab_data = file_obj.read( shdr.size ).decode() # this is used as string!

  # Load sections

  symtab_data = None
  strtab_data = None

  mem_image = SparseMemoryImage()

  for section_idx in range(ehdr.shnum):

    # Read the data for the section header

    file_obj.seek( ehdr.shoff + section_idx * ehdr.shentsize )
    shdr_data = file_obj.read(ehdr.shentsize)

    # Pad the returned string in case the section header is not long
    # enough (otherwise the unpack function would not work)

    shdr_data = shdr_data.ljust( ElfSectionHeader.NBYTES, b'\0' )

    # Construct a section header object

    shdr = ElfSectionHeader( shdr_data )

    # Find the section name

    start = shstrtab_data[shdr.name:]
    section_name = start.partition('\0')[0]

    # Only sections marked as alloc should be written to memory

    if not ( shdr.flags & ElfSectionHeader.FLAGS_ALLOC ):
      continue

    # Read the section data if it exists

    if section_name not in ['.sbss', '.bss']:
      file_obj.seek( shdr.offset )
      data = file_obj.read( shdr.size )

    # NOTE: the .bss and .sbss sections don't actually contain any
    # data in the ELF.  These sections should be initialized to zero.
    # For more information see:
    #
    # - http://stackoverflow.com/questions/610682/bss-section-in-elf-file

    else:
      data = b'\0' * shdr.size

    # Save the data holding the symbol string table

    if shdr.type == ElfSectionHeader.TYPE_STRTAB:
      strtab_data = data

    # Save the data holding the symbol table

    elif shdr.type == ElfSectionHeader.TYPE_SYMTAB:
      symtab_data = data

    # Otherwise create section and append it to our list of sections

    else:
      section = SparseMemoryImage.Section( section_name, shdr.addr, data )
      mem_image.add_section( section )

  # Load symbols. We skip the first symbol since it both "designates the
  # first entry in the table and serves as the undefined symbol index".
  # For now, I have commented this out, since we are not really using it.

  # num_symbols = len(symtab_data) / ElfSymTabEntry.NBYTES
  # for sym_idx in xrange(1,num_symbols):
  #
  #   # Read the data for a symbol table entry
  #
  #   start = sym_idx * ElfSymTabEntry.NBYTES
  #   sym_data = symtab_data[start:start+ElfSymTabEntry.NBYTES]
  #
  #   # Construct a symbol table entry
  #
  #   sym = ElfSymTabEntry( sym_data )
  #
  #   # Get the symbol type
  #
  #   sym_type  = sym.info & 0xf
  #
  #   # Check to see if symbol is one of the three types we want to load
  #
  #   valid_sym_types = \
  #   [
  #     ElfSymTabEntry.TYPE_NOTYPE,
  #     ElfSymTabEntry.TYPE_OBJECT,
  #     ElfSymTabEntry.TYPE_FUNC,
  #   ]
  #
  #   # Check to see if symbol is one of the three types we want to load
  #
  #   if sym_type not in valid_sym_types:
  #     continue
  #
  #   # Get the symbol name from the string table
  #
  #   start = strtab_data[sym.name:]
  #   name = start.partition('\0')[0]
  #
  #   # Add symbol to the sparse memory image
  #
  #   mem_image.add_symbol( name, sym.value )

  return mem_image

#-------------------------------------------------------------------------
# elf_writer
#-------------------------------------------------------------------------
# Writes a sparse memory image object to an ELF file. Currently we write
# the ELF file in the following order:
#
#  - ElfHeader
#  - ElfSectionHeader for "null" section
#  - ElfSectionHeader for all "normal" sections
#  - ElfSectionHeader for ".shstrtab" section
#  - data for all "normal" sections
#  - data for ".shstrtab" section
#

def elf_writer( mem_image, file_obj ):

  # Get the sections

  sections = mem_image.get_sections()

  ehdr = ElfHeader()

  # Many of these fields are just copied from what binutils generates.
  # Note that we have two extra sections beyond the normal sections. The
  # first "null" section and the final ".shstrtab" section.

  ehdr.ident     = "\x7fELF\x01\x01\x01".ljust( ElfHeader.IDENT_NBYTES, '0' )
  ehdr.type      = ElfHeader.TYPE_EXEC
  ehdr.machine   = 8
  ehdr.version   = 1
  ehdr.entry     = 0x00001000
  ehdr.phoff     = 0
  ehdr.shoff     = ElfHeader.NBYTES         # shdrs right after ehdr
  ehdr.flags     = 0x70b03000
  ehdr.ehsize    = 0
  ehdr.phentsize = 0
  ehdr.phnum     = 0
  ehdr.shentsize = ElfSectionHeader.NBYTES  # shdrs are fixed size
  ehdr.shnum     = len(sections) + 2        # add 2 for extra sections
  ehdr.shstrndx  = len(sections) + 1        # location of shstrtab

  # Write the ELF header to the file

  file_obj.write( ehdr.to_bytes() )

  # Write the first "null" section header to the file

  shdr = ElfSectionHeader()
  shdr.name      = 0
  shdr.type      = 0
  shdr.flags     = 0
  shdr.addr      = 0
  shdr.offset    = 0
  shdr.size      = 0
  shdr.link      = 0
  shdr.info      = 0
  shdr.addralign = 0
  shdr.entsize   = 0
  file_obj.write( shdr.to_bytes() )

  # The section data is going to start after the ELF header and all of
  # the section headers.

  section_offset =  ElfHeader.NBYTES                        # ELF header
  section_offset += ElfSectionHeader.NBYTES                 # null shdr
  section_offset += len(sections) * ElfSectionHeader.NBYTES # normal shdrs
  section_offset += 1 * ElfSectionHeader.NBYTES             # shstrtab shdr

  # Collect section names in a string for writing to ".shstrtab"

  section_names = "\0"

  # Write the "normal" section headers to the file

  for section in sections:

    shdr = ElfSectionHeader()
    shdr.name      = len(section_names)
    shdr.type      = ElfSectionHeader.TYPE_PROGBITS
    shdr.flags     = ElfSectionHeader.FLAGS_ALLOC
    shdr.addr      = section.addr
    shdr.offset    = section_offset
    shdr.size      = len(section.data)
    shdr.link      = 0
    shdr.info      = 0
    shdr.addralign = 0
    shdr.entsize   = 0

    file_obj.write( shdr.to_bytes() )

    section_names  += section.name + "\0"
    section_offset += len(section.data)

  # Write the ".shstrtab" section header to the file

  shdr = ElfSectionHeader()
  shdr.name      = len(section_names)
  shdr.type      = ElfSectionHeader.TYPE_STRTAB
  shdr.flags     = 0
  shdr.addr      = 0
  shdr.offset    = section_offset
  shdr.size      = len( section_names + ".shstrtab\0" )
  shdr.link      = 0
  shdr.info      = 0
  shdr.addralign = 0
  shdr.entsize   = 0

  file_obj.write( shdr.to_bytes() )

  section_names  += ".shstrtab\0"
  section_offset += len(section_names)

  # Write the section data for "normal" sections

  for section in sections:
    file_obj.write( section.data )

  # Write the data for the ".shstrtab" section

  file_obj.write( section_names.encode() )



"""
==========================================================================
 test_masters.py
==========================================================================

 Author: Shunning Jiang
   Date: May 27, 2020
"""
from pymtl3 import *
from pymtl3.stdlib.ifcs import MasterIfcCL

from .test_sinks import TestSinkCL
from .test_srcs import TestSrcCL


class TestMasterCL( Component ):

  def construct( s, ReqType, RespType, MasterIfc=MasterIfcCL ):
    s.master = MasterIfc( ReqType, RespType )
    s.src  = TestSrcCL( ReqType )
    s.sink = TestSinkCL( RespType )

    s.src.send  //= s.master.req
    s.sink.recv //= s.master.resp

  def done( s ):
    return s.src.done() and s.sink.done()

  def line_trace( s ):
    return str(s.master)



#=========================================================================
# YosysTranslationPass.py
#=========================================================================
# Author : Peitian Pan
# Date   : June 8, 2019
"""Translate a PyMTL component hierarhcy into yosys-verilog source code."""

from pymtl3.passes.backends.verilog import VerilogTranslationPass

from .YosysTranslator import YosysTranslator


class YosysTranslationPass( VerilogTranslationPass ):

  def __call__( s, top ):
    s.top = top
    s.translator = YosysTranslator( s.top )
    s.traverse_hierarchy( top )



"""
==========================================================================
 test_masters.py
==========================================================================

 Author: Shunning Jiang
   Date: May 27, 2020
"""
from pymtl3 import *
from pymtl3.stdlib.ifcs.valrdy_master_minion_ifcs import MasterIfcRTL

from .valrdy_test_sinks import TestSinkRTL
from .valrdy_test_srcs import TestSrcRTL


class TestMasterRTL( Component ):

  def construct( s, ReqType, RespType, MasterIfc=MasterIfcRTL ):
    s.master = MasterIfc( ReqType, RespType )
    s.src  = TestSrcRTL( ReqType )
    s.sink = TestSinkRTL( RespType )

    s.src.out  //= s.master.req
    s.sink.in_ //= s.master.resp

  def done( s ):
    return s.src.done() and s.sink.done()

  def line_trace( s ):
    return str(s.master)
